from typing import Mapping

from twisted.web.server import Request

from canonical.txapi.server.schema import List, Schema, Unicode
from canonical.landscape.api.helpers import (
    with_computer_selection,
    with_computer,
)
from canonical.landscape.api.principal import Principal
from canonical.landscape.api.schema import Array
from canonical.landscape.api.types import ApiResponse
from canonical.landscape.model.computer.management import ComputerManagement
from canonical.landscape.model.main.account import Account, get_account_by_id
from canonical.landscape.model.main.computer import Computer, get_computer_by_id
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.model.users.users import UserSelection, UserManagement


UPDATE_ACTIONS = ("add", "remove")


USER_GROUP_UPDATE_BULK_SCHEMA = Schema(
    name="UpdateUserGroupsBulk",
    parameters=[
        Unicode("action", doc=f"Must be one of {UPDATE_ACTIONS}"),
        Array("usernames", Unicode()),
        Array("groupnames", Unicode()),
    ],
    doc="Add/remove every provided username from every provided groupname on the computer",
)


COMPUTER_GROUPS_LIST_SCHEMA = Schema(
    name="ComputerGroupsList", doc="Get all user groups for the provided computer_id"
)


COMPUTER_USERS_GROUPS_LIST_SCHEMA = Schema(
    name="ComputerUsersGroupsList",
    doc="Get all the groups for the provided username on the given computer_id",
)


@with_computer(COMPUTER_GROUPS_LIST_SCHEMA, permission="ViewComputer")
async def computer_groups_list_handler(
    request: Request,
    principal: Principal,
    computer: Computer,
    selection: ComputerSelection,
    validated_data: dict,
    url_args: Mapping[str, int],
) -> ApiResponse:
    """
    Get all user groups for the provided computer_id
    """
    groups = computer.get_groups()
    return {"groups": [group.serialize() for group in groups]}, 200


@with_computer(COMPUTER_USERS_GROUPS_LIST_SCHEMA, permission="ViewComputer")
async def computer_user_groups_list_handler(
    request: Request,
    principal: Principal,
    computer: Computer,
    selection: ComputerSelection,
    validated_data: dict,
    url_args: Mapping[str, int],
) -> ApiResponse:
    """
    Get all groups for a username on the provided computer_id
    """
    username = url_args["username"]
    user = computer.get_user_by_username(username)
    if not user:
        return {
            "error": "UnknownComputerUserError",
            "message": f"No user '{username}' on computer with id {computer.id}",
        }, 404
    return {"groups": [group.serialize() for group in user.groups]}, 200


@with_computer_selection(USER_GROUP_UPDATE_BULK_SCHEMA, "EditUser")
async def computer_usergroups_update_bulk_handler(
    request: Request,
    principal: Principal,
    selection: ComputerSelection,
    validated_data: dict,
    url_args: Mapping[str, int],
) -> ApiResponse:
    """
    Add/remove every provided username from every provided groupname on the computer
    """
    action = validated_data["action"]
    if action not in UPDATE_ACTIONS:
        return {
            "error": "InvalidParameterValueError",
            "message": f"action must be one of {UPDATE_ACTIONS}",
        }, 400
    usernames = set(validated_data["usernames"])
    groupnames = set(validated_data["groupnames"])

    person = get_person_by_id(principal.id)
    account = get_account_by_id(principal.account_id)
    computer_id = url_args["computer_id"]
    computer = account.get_computer(computer_id)

    if not computer:
        return {
            "error": "UnknownComputerError",
            "message": f"Invalid computer_id: {computer_id}",
        }, 400

    users = list(computer.get_users_by_usernames(usernames))
    found_usernames = set(user.username for user in users)
    if found_usernames != usernames:
        return {
            "error": "UnknownUserError",
            "message": f"No user(s) found with username(s) {usernames - found_usernames}",
        }, 400

    management = UserManagement(selection, person)
    user_selection = UserSelection(management, users)

    if action == "add":
        activity_group = user_selection.add_to_groups(groupnames)
    elif action == "remove":
        activity_group = user_selection.remove_from_groups(groupnames)

    if not activity_group:
        return {}, 200

    return activity_group.get_state(), 202
