from typing import Mapping

from twisted.web.server import Request
from canonical.txapi.server.schema import Schema, Unicode, List, UnicodeLine, Integer

from canonical.landscape.api.error import MissingParameterError
from canonical.landscape.api.principal import Principal
from canonical.landscape.api.helpers import (
    with_schema,
    with_management_class,
    paginated,
    EMPTY_BODY,
)
from canonical.landscape.api.schema import Array
from canonical.landscape.model.access.errors import UnknownRoleError
from canonical.landscape.model.main.schema import InvalidEmailAddress
from canonical.landscape.model.account.errors import (
    MaximumAccountAdministratorsError,
    IllegalInvitationError,
    RepeatedInvitationError,
)
from canonical.landscape.model.account.invitation import (
    InvitationManagement,
)


POST_INVITATION_SCHEMA = Schema(
    name="PostInvitation",
    parameters=[
        UnicodeLine("name", doc="The name of the person to invite."),
        UnicodeLine(
            "email",
            doc="The email address of the administrator, to "
            "which the invitation will be send.",
        ),
        Array(
            "roles",
            Unicode(),
            optional=True,
            default=["GlobalAdmin"],
            doc="If specified, the roles that the administrator is "
            "going to have in your account. "
            "Default to GlobalAdmin",
        ),
    ],
    doc="""Create and send an invitation to a new administrator for your account.""",
)


@with_management_class(InvitationManagement, permissions=["InviteAdministrator"])
async def invitation_get_handler(
    request: Request,
    principal: Principal,
    management: InvitationManagement,
    url_args=None,
):
    """Get an invitation with the given id in the account associated with the principal"""
    invitation_id = url_args["id"]
    invitation = management.get_invitation_by_id(invitation_id)
    if invitation is None:
        return {
            "error": "InvitationNotFound",
            "message": f"No invitation with id {invitation_id}",
        }, 404
    else:
        return invitation.serialize(), 200


@with_management_class(InvitationManagement, permissions=["InviteAdministrator"])
@paginated
async def invitation_list_handler(
    request: Request,
    principal: Principal,
    management: InvitationManagement,
    offset=None,
    limit=None,
):
    """Get all invitations for the account that the principal is associated with"""
    invitations = management.get_invitations()
    count = invitations.count()
    invitations.config(offset=offset, limit=limit)
    results = [invitation.serialize() for invitation in invitations]
    return {
        "count": count,
        "results": results,
    }, 200


@with_management_class(InvitationManagement, permissions=["InviteAdministrator"])
@with_schema(POST_INVITATION_SCHEMA)
async def invitation_post_handler(
    request: Request,
    principal: Principal,
    management: InvitationManagement,
    validated_data=Mapping[str, str],
):
    """Create a new invitation to the account associated with the principal for the
    given email.  Delete an existing invitation for that email, if present."""
    name = validated_data["name"]
    email = validated_data["email"]
    roles = validated_data["roles"]
    try:
        invitation = management.create_or_renew_invitation(name, email, roles)
    except (
        UnknownRoleError,
        MissingParameterError,
        MaximumAccountAdministratorsError,
        InvalidEmailAddress,
        IllegalInvitationError,
        RepeatedInvitationError,
    ) as e:
        return {"error": f"{e.__class__.__name__}", "message": f"{e}"}, 400
    else:
        try:
            invitation.send_invitation()
        except:
            return {
                "error": "LandscapeMailerError",
                "message": "Failed to send invitation email.",
            }, 500
        return {"id": invitation.id, "secure_id": invitation.secure_id}, 200


@with_management_class(InvitationManagement, permissions=["InviteAdministrator"])
async def invitation_delete_handler(
    request: Request,
    principal: Principal,
    management: InvitationManagement,
    url_args=None,
):
    """Delete an invitation with the given id in the account associated with the
    principal"""
    invitation_id = url_args["id"]
    invitation = management.get_invitation_by_id(invitation_id)
    if invitation:
        invitation.cancel()
        return EMPTY_BODY, 204
    else:
        return {
            "error": "InvitationNotFound",
            "message": f"No invitation with id {invitation_id}",
        }, 404
