import datetime
import logging
from functools import wraps
from typing import Callable, Optional

import jwt

from twisted.web.server import Request

from zope.component import getUtility

from canonical.landscape.api.types import RequestHandler
from canonical.landscape.api.principal import AnonymousCredentials, Principal
from canonical.landscape.interfaces import IFileBasedConfiguration
from canonical.landscape.model.configuration import is_development_mode
from canonical.landscape.model.main.account import Account, get_account_by_name
from canonical.landscape.model.main.person import (
    Person,
    get_person_by_id,
)

JWT_ALGO = "HS256"
JWT_EXPIRATION_TIME = datetime.timedelta(days=1)


class JwtException(Exception):
    """An exception originating from the decoding of a JWT."""


class JwtExpiredException(JwtException):
    """An exception originating from decoding an expired JWT."""

    code = "AuthTokenExpired"
    message = "Auth token expired. Please log in again."


class JwtInvalidException(JwtException):
    """An exception originating from decoding an invalid JWT."""

    code = "AuthTokenInvalid"
    message = "Auth token invalid. Please log in again."


class JwtInvalidSecret(JwtException):
    """An exception originating from an invalid JWT configuration."""

    code = "ServerConfigurationInvalid"
    message = "JWT secret not configured properly."


def encode_auth_token(
    person: Person, account: Optional[Account], secret_token: bytes
) -> str:
    """Creates a JWT for `person`, associated only with the given `account`.

    :param Person person: The person with are creating a token for.
    :param account: The account this JWT should be associated with.
    :type account: Account or None
    :param bytes secret_token: Server-side token used for encryption.

    :returns str: The JWT.
    """
    if account is None:
        account = person.get_default_account()

    now = datetime.datetime.utcnow()
    payload = {
        "exp": now + JWT_EXPIRATION_TIME,
        "iat": now,
        "sub": person.email,
        "acc": account.name,
        "id": person.id,
    }
    token = jwt_encode(payload, secret_token, algorithm=JWT_ALGO)

    if isinstance(token, bytes):
        return token.decode()

    return token


def get_secret_token() -> bytes:
    """Retrieves the secret token for JWT encryption from the service configuration, providing a
    default (and logging a warning) if one is not set.

    :returns bytes: The secret token.
    """
    config = getUtility(IFileBasedConfiguration)

    try:
        token = config.get("landscape")["secret-token"]
    except KeyError:
        if is_development_mode():
            logging.warn("FIX: no secret-token set, using default")
            return b"LANDSCAPESECRETTOKENDONOTUSE"
        else:
            logging.error("secret-token must be set in service.conf")
            raise JwtInvalidSecret

    return token.encode()


def jwt_authenticate_request(request: Request) -> Principal:
    """Decodes a JWT provided in `request` Authorization header and translates it to an
    authenticated Principal.

    :param Request request: The request to authenticate.

    :returns Principal: The authenticated user.
    """
    auth_header = request.getHeader("Authorization")

    if isinstance(auth_header, bytes):
        auth_header = auth_header.decode()

    if not auth_header or not auth_header.startswith("Bearer "):
        raise JwtInvalidException()

    auth_token = auth_header.split(" ")[1]
    secret_token = get_secret_token()

    try:
        payload = jwt.decode(auth_token, secret_token, algorithms=JWT_ALGO)

        person_id = payload.get("id")
        account_name = payload.get("acc")
        if person_id is None or account_name is None:
            raise JwtInvalidException()

        person = get_person_by_id(person_id)
        account = get_account_by_name(account_name)
        if not person or not account:
            raise JwtInvalidException()

        credentials = person.get_api_credentials(account)
        if credentials is None:
            credentials = person.create_api_credentials(account)

        return Principal(credentials)
    except jwt.ExpiredSignatureError:
        raise JwtExpiredException()
    except jwt.InvalidTokenError:
        raise JwtInvalidException()


def login_required(handler: Callable) -> RequestHandler:
    """Wraps `handler` in a layer of JWT authentication, adding the authenticated `Principal` to the
    handlers parameters. This decorator should be the first one applied to a handler.

    :param Callable handler: The handler to wrap.

    :returns RequestHandler: The wrapped handler, ready to handle the request.
    """

    @wraps(handler)
    async def wrapper(request, *args, **kwargs):
        try:
            principal = jwt_authenticate_request(request)
            return await handler(request, principal, *args, **kwargs)
        except JwtException as exc:
            return {"error": exc.code, "message": exc.message}, 401

    return wrapper


def jwt_encode(*args, **kwargs) -> str:
    """Wrapper for `jwt.encode` that ensures the result is a `str` across versions of the lib."""
    token = jwt.encode(*args, **kwargs)

    if isinstance(token, bytes):
        return token.decode()

    return token
