# Copyright 2005-2013 Canonical Limited.  All rights reserved.

"""
Packages
========

The methods available here are related to package management operations.
"""

from collections import defaultdict
from typing import Optional

from zope.security.interfaces import Unauthorized
from twisted.web.http import Request

from canonical.txapi.server.schema import (
    Bool,
    Integer,
    List,
    PositiveInteger,
    Schema,
    Unicode,
)
from canonical.txapi.server.method import method

from canonical.lib.timestamp import parse_datetime
from canonical.search.engine import SearchCriteriaError

from canonical.landscape.api.error import (
    InvalidTimeError,
    UnknownPackagesError,
    InvalidQueryError,
    UnknownComputersError,
)
from canonical.landscape.api import RESULT_OFFSET, RESULT_LIMIT
from canonical.landscape.api.helpers import (
    paginated,
    with_schema,
    with_computer,
    ApiResponse,
    Principal,
)
from canonical.landscape.api.jwt_auth import login_required
from canonical.landscape.api.methods import AccountMethod
from canonical.landscape.api.schema import Array
from canonical.landscape.model.activity.schedule import MAX_SCHEDULE_AFTER_WINDOW
from canonical.landscape.model.computer.collection import ComputerCollection
from canonical.landscape.model.package.collection import (
    parse_package_names,
    PackageCollection,
)
from canonical.landscape.model.package.management import PackageManagement
from canonical.landscape.model.package.search import PackageSearch, DOWNGRADE
from canonical.landscape.model.main.account import get_account_by_id
from canonical.landscape.model.main.computer import Computer
from canonical.landscape.model.main.person import get_person_by_id
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.api.routes import SelectionRoute

__all__ = [
    "GetPackages",
    "InstallPackages",
    "RemovePackages",
    "UpgradePackages",
    "packages_get_handler",
    "computer_packages_list_handler",
    "computer_package_downgrades_get_handler",
    "computer_packages_install_at_version_post_handler",
]


def get_deliver_after_time(deliver_after):
    """Returns a datetime for the time an activity should be scheduled for.

    @param deliver_after: String with the time to be converted.
    @raises InvalidTimeError: If the string provided cannot be converted.
    @return datetime or None if the deliver_after string could not be
        converted.
    """
    deliver_after_time = None
    if deliver_after:
        try:
            deliver_after_time = parse_datetime(deliver_after)
        except ValueError:
            raise InvalidTimeError(
                "Invalid format for deliver_after: '%s'" % deliver_after
            )
    return deliver_after_time


def _serialize_package_states(package_states):
    result = []
    for package, states in package_states:
        package_info = package.get_state()
        computers_info = {
            "available": [],
            "installed": [],
            "upgrades": [],
            "held": [],
        }

        for available, installed, held, upgrade, computer_ids in states:
            if available:
                computers_info["available"].extend(computer_ids)
            if installed:
                computers_info["installed"].extend(computer_ids)
            if held:
                computers_info["held"].extend(computer_ids)
            if upgrade:
                computers_info["upgrades"].extend(computer_ids)

        package_info["computers"] = computers_info
        result.append(package_info)
    return result


@method
class GetPackages(AccountMethod):
    """
    Get a list of packages associated with the account used for
    authentication.

    - **query**: A query string used to select computers to query packages on.
      See :ref:`Computer Queries <computer-query>` for details.
    - **search**: A string to restrict the search to (optional).  All
      fields are searched, not just those returned. (e.g., description)
    - **names**: Restrict the search to these package names. Multiple
      names can be specified by numbering the names with names.1,
      names.2, etc.
    - **installed**: If true only packages in the installed state will be
      returned, if false only packages not installed will be returned. If not
      given both installed and not installed packages will be returned.
    - **available**: If true only packages in the available state will be
      returned, if false only packages not available will be returned. If not
      given both available and not available packages will be returned.
    - **upgrade**: If true, only installable packages that are upgrades
      for an installed one are returned. If false, only installable packages
      that are not upgrades are returned. If not given, packages will be
      returned regardless of wether they are upgrades or not.
    - **held**: If true, only installed packages that are held on computers are
      returned. If false, only packages that are not held on computers are
      returned. If not given, packages will be returned regardless of the held
      state.
    - **limit**: The maximum number of results returned by the method. It
      defaults to 1000.
    - **offset**: The offset inside the list of results.

    A package is considered installed if dpkg reports it as installed on the
    system.

    A package is considered available if it can be fetched from an APT
    source. Note that this means that it's possible for an installed package
    to be not available.

    A package is considered an upgrade if it's available and if it has a
    version higher than a non-held installed package with the same name.

    For example, the following request looks for a package named with the tag
    'server' with a limit of 20 packages::

        ?action=GetPackages&query=tag:server&search=python&limit=20

    The method returns a JSON serialized list of packages, with the list of
    computer IDs on which they are available, installed, or available as
    upgrades::

        [{"name": "python2.7"
          "summary": "An interactive high-level object-oriented language...",
          "computers": {
              "available": [12, 17], "installed": [], "upgrades": [],
              "held": []},
          "version": "2.7.2"},
         {"name": "python2.6"
          "summary": "An interactive high-level object-oriented language...",
          "computers": {
              "available": [12, 17], "installed": [12, 17], "upgrades": [],
              "held": [12]},
          "version": "2.6.5"}]

    """

    requires = "ViewComputerPackages"

    schema = Schema(
        name="GetPackages",
        parameters=[
            Unicode(
                "query",
                doc="A query string used to select computers to query packages on.",
            ),
            Unicode(
                "search",
                optional=True,
                doc=(
                    "A string to restrict the search to.  All fields are "
                    "searched, not just those returned. "
                    "(e.g., description)"
                ),
            ),
            List(
                "names",
                Unicode(),
                optional=True,
                doc="Restrict the search to these package names.",
            ),
            Bool(
                "installed",
                optional=True,
                doc=(
                    "If true only packages in the installed state will "
                    "be returned, if false only packages not installed "
                    "will be returned. If not given both installed and "
                    "not installed packages will be returned."
                ),
            ),
            Bool(
                "available",
                optional=True,
                doc=(
                    "If true only packages in the available state will "
                    "be returned, if false only packages not available "
                    "will be returned. If not given both available and "
                    "not available packages will be returned."
                ),
            ),
            Bool(
                "upgrade",
                optional=True,
                doc=(
                    "If true, only installable packages that are "
                    "upgrades for an for an installed one are returned. "
                    "If false, only installable packages that are not "
                    "upgrades are returned. If not given, packages will "
                    "be returned regardless of wether they are upgrades "
                    "or not."
                ),
            ),
            Bool(
                "held",
                optional=True,
                doc=(
                    "If true, only installed packages that are held on "
                    "computers are returned. If false, only packages "
                    "that are not held on computers are returned. If "
                    "not given, packages will be returned regardless "
                    "of the held state."
                ),
            ),
            Integer(
                "offset",
                default=RESULT_OFFSET,
                optional=True,
                doc="The offset inside the list of results.",
            ),
            PositiveInteger(
                "limit",
                default=RESULT_LIMIT,
                optional=True,
                doc=(
                    "The maximum number of results returned by the "
                    "method. It defaults to 1000."
                ),
            ),
        ],
        errors=[InvalidQueryError, UnknownComputersError],
        doc="""
    Get a list of packages associated with the account used for
    authentication.

    A package is considered installed if dpkg reports it as installed on the
    system.

    A package is considered available if it can be fetched from an APT
    source. Note that this means that it's possible for an installed package
    to be not available.

    A package is considered an upgrade if it's available and if it has a
    version higher than a non-held installed package with the same name.
    """,
    )

    route = SelectionRoute(permission=requires, enforce_limits=False)

    def query(self, context, call):
        """
        Find all computers matching query and query the packages on those
        computers.
        """
        names = call.args.names if call.args.names else None
        management = PackageSearch(context)
        package_states, _ = management.get_package_states(
            text=call.args.search,
            names=names,
            limit=call.args.limit,
            offset=call.args.offset,
            installed=call.args.installed,
            available=call.args.available,
            upgrades=call.args.upgrade,
            held=call.args.held,
        )

        result = _serialize_package_states(package_states)

        return result


PACKAGES_GET_HANDLER_SCHEMA = GetPackages.schema.extend()
PACKAGES_GET_HANDLER_SCHEMA.replace_parameter(
    "names",
    Array(
        "names",
        Unicode(),
        optional=True,
        doc=("Restrict the search to these package names."),
    ),
)


@login_required
@with_schema(PACKAGES_GET_HANDLER_SCHEMA)
@paginated
async def packages_get_handler(
    request: Request,
    principal: Principal,
    validated_data: dict,
    search_args: Optional[dict] = None,
    limit: int = RESULT_LIMIT,
    offset: int = RESULT_OFFSET,
):
    permission = "landscape." + GetPackages.requires

    collection = ComputerCollection(
        account_id=principal.account_id,
        person_id=principal.id,
        permission=permission,
    )
    query = validated_data["query"]

    try:
        collection = collection.with_search_criteria(query)
    except SearchCriteriaError as error:
        return {
            "error": "InvalidQueryError",
            "message": f"Invalid query '{error}'.",
        }, 400

    result = collection.find()
    result.order_by(Computer.title)
    computers = list(result)

    account = get_account_by_id(principal.account_id)
    person = get_person_by_id(principal.id)
    selection = ComputerSelection(account, person=person, permission=permission)
    selection.select_computers(computers)

    if not selection.get_selected_count():
        return {
            "error": "UnknownComputersError",
            "message": f"No computer(s) matching the query '{query}' and "
            "authorised for this operation.",
        }, 400

    management = PackageSearch(selection)
    package_states, package_count = management.get_package_states(
        text=validated_data.get("search"),
        names=validated_data.get("names") or None,
        limit=validated_data.get("limit"),
        offset=validated_data.get("offset"),
        installed=validated_data.get("installed"),
        available=validated_data.get("available"),
        upgrades=validated_data.get("upgrade"),
        held=validated_data.get("held"),
        count_results=True,
    )

    result = _serialize_package_states(package_states)

    return {
        "count": package_count,
        "results": result,
    }, 200


computer_packages_list_schema = Schema(
    name="ComputerPackagesList",
    parameters=[
        Unicode(
            "search",
            optional=True,
            doc=(
                "A string to restrict the search to. The names and description"
                "fields are searched."
            ),
        ),
        Array(
            "names",
            Unicode(),
            optional=True,
            doc="Restrict the search to these package names.",
        ),
        Bool(
            "installed",
            optional=True,
            doc=(
                "If true, only packages in the installed state will be "
                "returned. If false, only packages not installed will be "
                "returned. If not given, both installed and not installed "
                "packages will be returned."
            ),
        ),
        Bool(
            "available",
            optional=True,
            doc=(
                "If true, only packages in the available state will be "
                "returned. If false, only packages not available will be "
                "returned. If not given, both available and not available "
                "packages will be returned."
            ),
        ),
        Bool(
            "upgrade",
            optional=True,
            doc=(
                "If true, only installable packages that are upgrades for an "
                "installed one are returned. If false, only installable "
                "packages that are not upgrades are returned. If not given, "
                "packages will be returned regardless of whether they are "
                "upgrades or not."
            ),
        ),
        Bool(
            "held",
            optional=True,
            doc=(
                "If true, only installed packages that are held on computers "
                "are returned. If false, only packages that are not held on "
                "computers are returned. If not given, packages will be "
                "returned regardless of the held state."
            ),
        ),
        Bool(
            "security",
            optional=True,
            doc=(
                "If true, only installable packages that are security upgrades "
                "for an installed one are returned. If false, only installable "
                "packages that are not security upgrades are returned. If not "
                "given, packages will be returned regardless of whether they "
                "are security upgrades or not."
            ),
        ),
        Integer(
            "offset",
            default=RESULT_OFFSET,
            optional=True,
            doc="The offset inside the list of results.",
        ),
        PositiveInteger(
            "limit",
            default=RESULT_LIMIT,
            optional=True,
            doc=(
                "The maximum number of results returned by the method. It "
                "defaults to 1000."
            ),
        ),
    ],
    errors=[UnknownComputersError],
    doc="""
    Get a list of packages associated with the account used for
    authentication.

    A package is considered installed if dpkg reports it as installed on the
    system.

    A package is considered available if it is not installed or held.

    A package is considered an upgrade if it's available and if it has a
    version higher than an installed package with the same name.
    """,
)


@login_required
@with_schema(computer_packages_list_schema)
@paginated
async def computer_packages_list_handler(
    request: Request,
    principal: Principal,
    url_args: dict,
    validated_data: dict,
    limit: int = RESULT_LIMIT,
    offset: int = RESULT_OFFSET,
) -> ApiResponse:
    account = get_account_by_id(principal.account_id)
    person = get_person_by_id(principal.id)
    permission = "landscape." + GetPackages.requires
    selection = ComputerSelection(account, person=person, permission=permission)
    computer_id = url_args.get("computer_id")
    selection.select_computers_by_id([computer_id])

    if not selection.get_selected_count():
        return {
            "error": "UnknownComputersError",
            "message": "Computer does not exist or is not authorised for this operation",
        }, 400

    text = validated_data.get("search")
    names = validated_data.get("names")
    installed = validated_data.get("installed")
    available = validated_data.get("available")
    upgrade = validated_data.get("upgrade")
    held = validated_data.get("held")
    security = validated_data.get("security")
    limit = limit
    offset = offset
    management = PackageSearch(selection)

    if any([upgrade, held, security]) and not all([available, installed]):
        # Use the query that omits the use of available packages if possible
        # since the list of package ids tends to be very large.
        packages = management.get_nonavailable_computer_package_states(
            text, names, upgrade, held, security, limit, offset
        ).get(computer_id, [])
    else:
        packages = management.get_computer_package_states(
            text, names, available, installed, upgrade, held, security, limit, offset
        ).get(computer_id, [])

    return packages, 200


class BasicPackageOperation(AccountMethod):
    """
    Base class for package operations.

    Override the L{compute_changes} method in subclasses in order to return the
    correct package mapping.
    """

    requires = "EditComputerPackages"

    schema = Schema(
        name="BasicPackageOperation",
        parameters=[
            Unicode("query", doc=("A qualified criteria to be used in the search.")),
            List(
                "packages",
                Unicode(),
                doc=(
                    "A list of package names on which to operate. "
                    "Multiple package names can be supplied."
                ),
            ),
            Unicode(
                "deliver_after",
                optional=True,
                doc=("A time in the future to perform the package " "operation."),
            ),
            Integer(
                "deliver_delay_window",
                optional=True,
                min=0,
                max=MAX_SCHEDULE_AFTER_WINDOW,
                default=0,
                doc=(
                    "Randomise delivery within the given time frame "
                    "specified in minutes"
                ),
            ),
        ],
        errors=[
            UnknownPackagesError,
            InvalidTimeError,
            UnknownComputersError,
            InvalidQueryError,
            Unauthorized,
        ],
    )

    route = SelectionRoute(permission=requires)

    def query(self, context, call):
        management = PackageManagement(context)
        search = management.get_search()
        deliver_after = get_deliver_after_time(call.args.deliver_after)
        deliver_delay_window = call.args.deliver_delay_window

        all_names = set(parse_package_names(call.args.packages))
        package_states, _ = search.get_package_states(names=call.args.packages)
        package_names = set(package.name for package, states in package_states)

        not_found_packages = all_names.difference(package_names)
        if not_found_packages:
            package_names = ", ".join(sorted(not_found_packages))
            raise UnknownPackagesError("No such package(s) '%s'." % package_names)

        changes = self.compute_changes(package_states)
        if not changes:
            return {}

        person = get_person_by_id(call.principal.id)
        activity = management.change_packages(
            person,
            changes,
            deliver_after=deliver_after,
            batch_size=100,
            deliver_after_window=deliver_delay_window,
        )
        return activity.get_state()

    def compute_changes(self, package_states):
        """
        Compute package changes for installation or removal.

        @param packages_states: A C{list} of tuples with packages and a list of
            states with computer ids for each state.
        @return: A C{list} with the computer, packages to install and packages
            to remove.
        """
        raise NotImplementedError()


@method
class InstallPackages(BasicPackageOperation):
    """
    Install packages on selected computers.

    Request the installation of named packages on a selection of computers.

    - **query**: A qualified criteria to be used in the search. See
      :ref:`Computer Queries <computer-query>` for details.
    - **packages.#**: The list of package names to be installed, multiple
      names can be supplied.
    - **deliver_after**: Package installs will only take place when the
      computer contacts Landscape after this time (optional).
    - **deliver_delay_window**: Randomise delivery within the given time
      frame specified in minutes (optional).

    Example of a valid request::

        ?action=InstallPackages&query=tag:server&packages.1=python&
            packages.2=postgresql-8.4

    The method returns a JSON serialized list of the parent activity created::

       {'activity_status': 'undelivered',
        'computer_id': None,
        'creation_time': '2011-07-13T00:53:56Z',
        'creator': {'email': 'person@example.com',
                    'id': 66808,
                    'name': 'Person'},
        'deliver_delay_window': 0,
        'id': 4,
        'parent_id': None,
        'summary': 'Install packages haskell-mode and python-mode',
        'type': 'ActivityGroup'}

    """

    schema = BasicPackageOperation.schema.extend(
        name="InstallPackages", doc="Install packages on selected computers."
    )

    def compute_changes(self, package_states):
        """
        Compute changes for package installation according to the following
        rules::

        Loop over each computer found and:
        1. Get available packages, but only the ones related to the
        search results.
        2. Get installed packages.
        3. Check if there are packages that were already installed
        and remove all versions of the package from the installable
        package list.
        """
        # map package names to a list of 3-tuples with package, computer ids
        # with package available, computer ids with package installed
        packages_by_name = defaultdict(list)
        for package, states in package_states:
            computer_ids_installed = list()
            computer_ids_available = list()

            for available, installed, _, _, computer_ids in states:
                if available:
                    computer_ids_available.extend(computer_ids)
                elif installed:
                    computer_ids_installed.extend(computer_ids)

            packages_by_name[package.name].append(
                (package, computer_ids_available, computer_ids_installed)
            )

        changes_by_computer_id = defaultdict(list)
        for package_name, packages in packages_by_name.items():
            computer_ids_install_found = set()

            for package, available_ids, installed_ids in packages:
                package_id = package.id
                computer_ids_install_found.update(installed_ids)
                # packages are sorted by version newest to oldest, so the first
                # package we found for each name is the one to install
                computer_ids_to_install = set(available_ids).difference(
                    computer_ids_install_found
                )
                computer_ids_install_found.update(computer_ids_to_install)

                for computer_id in computer_ids_to_install:
                    changes_by_computer_id[computer_id].append(package_id)
        return [
            (computer_id, installs, (), (), ())
            for computer_id, installs in changes_by_computer_id.items()
        ]


@method
class RemovePackages(BasicPackageOperation):
    """
    Remove packages on selected computers.

    Request the removal of named packages on a selection of computers.

    - **query**: A qualified criteria to be used in the search. See
      :ref:`Computer Queries <computer-query>` for details.
    - **packages.#**: The package names to be removed, multiple package
      names can be supplied.
    - **deliver_after**: The removal will only take place when the computer
      contacts Landscape after this time (optional).
    - **deliver_delay_window**: Randomise delivery within the given time
      frame specified in minutes (optional).

    Example of a valid request::

        ?action=RemovePackages&query=tag:server&packages.1=python&
            packages.2=postgresql-8.4

    The method returns a JSON serialized list of the activity created::

       {'activity_status': 'undelivered',
        'computer_id': None,
        'creation_time': '2011-07-13T00:53:56Z',
        'creator': {'email': 'person@example.com',
                    'id': 66808,
                    'name': 'Person'},
        'deliver_delay_window': 0,
        'id': 4,
        'parent_id': None,
        'summary': 'Remove packages haskell-mode and python-mode',
        'type': 'ActivityGroup'}

    """

    schema = BasicPackageOperation.schema.extend(
        name="RemovePackages", doc="Remove packages on selected computers."
    )

    def compute_changes(self, package_states):
        """
        Compute changes for package removal.

        Remove a package only if it's installed and matched by the search.
        """
        removals_by_computer_id = defaultdict(list)

        for package, states in package_states:
            package_id = package.id
            for _, installed, _, _, computer_ids in states:
                if not installed:
                    continue

                for computer_id in computer_ids:
                    removals_by_computer_id[computer_id].append(package_id)

        return [
            (computer_id, (), removals, (), ())
            for computer_id, removals in removals_by_computer_id.items()
        ]


@method
class UpgradePackages(AccountMethod):
    """
    Request upgrading of all packages identified as being upgradable, on all
    computers selected by query.

    - **query**: A qualified criteria to be used in the search. See
      :ref:`Computer Queries <computer-query>` for details.
    - **packages.#**: A string to restrict the upgraded packages to install
      (optional).  Multiple package names can be supplied.
    - **security_only**: If 'true' then only packages with USNs or from
      the security pocket, i.e. security upgrades will be applied (optional).
    - **deliver_after**: The upgrade will only take place when the computer
      contacts Landscape after this time (optional).
    - **deliver_delay_window**: Randomise delivery within the given time
      frame specified in minutes (optional).


    Example of a valid request::

        UpgradePackages?query=tag:server&packages.1=python

    The method returns a JSON serialized list of the activity created::

        {'computer_id': None,
         'creation_time': '2011-07-18T15:30:13Z',
         'creator': {'email': 'person@example.com', 'id': 8634,
                     'name': 'Person'},
         'deliver_delay_window': 0,
         'id': 4547,
         'parent_id': None,
         'summary': 'Upgrade packages libpackz and packz',
         'type': 'ActivityGroup'}
    """

    requires = "EditComputerPackages"

    schema = Schema(
        name="UpgradePackages",
        parameters=[
            Unicode("query", doc=("A qualified criteria to be used in the search.")),
            List(
                "packages",
                Unicode(),
                optional=True,
                doc=(
                    "List of package names on which to perform an "
                    "upgrade. Multiple package names can be supplied "
                    "like packages.1=foo and packages.2=bar."
                ),
            ),
            Bool(
                "security_only",
                optional=True,
                default=False,
                doc=(
                    "If 'true' then only packages with USNs, i.e. "
                    "security upgrades will be applied."
                ),
            ),
            Unicode(
                "deliver_after",
                optional=True,
                doc=("A time in the future to perform the package " "upgrade."),
            ),
            Integer(
                "deliver_delay_window",
                optional=True,
                min=0,
                max=MAX_SCHEDULE_AFTER_WINDOW,
                default=0,
                doc=(
                    "Randomise delivery within the given time frame "
                    "specified in minutes"
                ),
            ),
        ],
        errors=[InvalidTimeError, UnknownComputersError, InvalidQueryError],
        doc="""
    Request upgrading of all packages identified as being upgradable, on all
    computers selected by query.
    """,
    )

    route = SelectionRoute(permission=requires)

    def query(self, context, call):
        management = PackageManagement(context)
        deliver_after = get_deliver_after_time(call.args.deliver_after)
        deliver_delay_window = call.args.deliver_delay_window
        search = management.get_search()
        changes = self.compute_changes(search, call)
        if changes is None:
            return {}

        person = get_person_by_id(call.principal.id)
        activity = management.change_packages(
            person,
            changes,
            deliver_after=deliver_after,
            batch_size=100,
            deliver_after_window=deliver_delay_window,
        )
        return activity.get_state()

    def compute_changes(self, search, call):
        """
        Return a list of package changes representing upgrades to the selected
        computers, or None if none can be found.
        """
        package_names = call.args.packages or None
        changes = search.get_package_upgrades_by_computer_id(
            names=package_names, security_only=call.args.security_only
        )

        for computer, installs, removals, _, _ in changes:
            if installs or removals:
                break
        else:
            return None
        return changes


computer_packages_downgrades_schema = Schema(
    name="ComputerPackageDowngrades",
    doc="Get available downgrades, if any, for a given package name "
    "installed on a computer",
)


@with_computer(computer_packages_downgrades_schema, "ViewComputerPackages")
async def computer_package_downgrades_get_handler(
    request: Request,
    principal: Principal,
    computer: Computer,
    selection: ComputerSelection,
    url_args: dict,
    validated_data: dict,
) -> ApiResponse:

    package_name = url_args["package_name"]

    # Full-text search is the only interface to package search -- this may over-produce
    # results that we will further filter later
    search = PackageSearch(selection)
    package_search_result = search.get_relations_by_package_name(
        text=package_name, installed=True, limit=None
    )

    downgrades = []
    package_is_installed = False
    for package_relation in package_search_result["relations"]:

        # Filter out packages that may have been partial matched and are therefore
        # *not* an exact match to the package that was passed in the query.
        # E.g., ignore python3-distutils-extra if looking for downgrades
        # to python3-distutils
        if package_name != package_relation["name"]:
            continue
        else:
            package_is_installed = True

        for relation in package_relation["relations"]:

            # PackageSearch interface doesn't support returning only downgrades
            if relation["kind"] == DOWNGRADE:
                package_id = relation["downgrade"]
                package = package_search_result["packages"][package_id]
                downgrades.append(package)

    if not package_is_installed:
        return {
            "error": "InstalledPackageNotFound",
            "message": f"No installed package with name '{package_name}' "
            f"found on computer with id '{computer.id}'",
        }, 404

    return {"results": [package.get_state() for package in downgrades]}, 200


COMPUTER_PACKAGES_INSTALL_AT_VERSION_SCHEMA = Schema(
    name="ComputerPackagesInstallAtVersion",
    parameters=[
        Unicode(
            "package_name",
            doc=("The Debian package name to install."),
        ),
        Unicode(
            "package_version",
            doc=("The Debian version to install."),
        ),
        Unicode(
            "deliver_after",
            optional=True,
            doc=("A time in the future to perform the package upgrade."),
        ),
        Integer(
            "deliver_delay_window",
            optional=True,
            min=0,
            max=MAX_SCHEDULE_AFTER_WINDOW,
            default=0,
            doc=("Randomise delivery within the given time frame specified in minutes"),
        ),
    ],
    doc="Install a package at a specific version.",
)


@with_computer(COMPUTER_PACKAGES_INSTALL_AT_VERSION_SCHEMA, "EditComputerPackages")
async def computer_packages_install_at_version_post_handler(
    request: Request,
    principal: Principal,
    computer: Computer,
    selection: ComputerSelection,
    url_args: dict,
    validated_data: dict,
) -> ApiResponse:

    package_name = validated_data["package_name"]
    package_version = validated_data["package_version"]

    try:
        deliver_after = get_deliver_after_time(validated_data.get("deliver_after"))
    except InvalidTimeError as e:
        return {"error": e.__class__.__name__, "message": str(e)}, 400
    deliver_delay_window = validated_data["deliver_delay_window"]

    search = PackageSearch(selection)
    package_states, _ = search.get_package_states(names=[package_name])
    for package, _ in package_states:
        if package.version == package_version:
            break
    else:
        return {
            "error": "PackageNotFound",
            "message": (
                f"No package with name '{package_name}' "
                f"at version {package_version} "
                f"available on computer with id '{computer.id}'"
            ),
        }, 400

    computer_changes = [(computer.id, (package.id,), (), (), ())]
    management = PackageManagement(selection)
    activity_parent = management.change_packages(
        principal.person,
        computer_changes,
        deliver_after=deliver_after,
        deliver_after_window=deliver_delay_window,
    )
    return activity_parent.get_state(), 202
