# Copyright 2005-2023 Canonical Limited. All rights reserved.
"""
Secrets
=======

The methods available here give the ability to store and retrieve secrets.
"""

import os

from base64 import b64decode
from typing import Optional

from canonical.txapi.server.call import Call
from canonical.txapi.server.schema import Integer, List, Schema, Unicode
from canonical.txapi.server.method import method

from canonical.landscape.api.error import (
    InvalidParameterCombinationError,
    MissingParameterError,
    SecretsAuthenticationError,
    SecretsSealedError,
    SecretNotFoundError,
)
from canonical.landscape.api.helpers import Data
from canonical.landscape.api.methods import AccountMethod
from canonical.landscape.api.routes import AccountRoute
from canonical.landscape.api.secrets_client import get_secrets_client, SecretsClient
from canonical.landscape.model.main.account import Account
from canonical.landscape.model.main.computer import Computer


__all__ = ["CreateSecret", "GetSecret", "ListSecrets"]


@method
class CreateSecret(AccountMethod):
    """Creates a secret for a computer."""

    requires = "CreateSecrets"
    route = AccountRoute()

    schema = Schema(
        name="CreateSecret",
        parameters=[
            Unicode("data_id", doc="The name of the secret to be stored."),
            Data("data", doc="The text of the secret to be stored"),
            Unicode(
                "secrets_name",
                optional=True,
                doc="A unique identifier for the computer. Note that this is "
                "different from the computer_id",
            ),
            Unicode(
                "token",
                optional=True,
                doc="The authentication token to be passed to the secrets " "manager",
            ),
        ],
        errors=[
            MissingParameterError,
            SecretNotFoundError,
            SecretsAuthenticationError,
            SecretsSealedError,
        ],
        doc="Stores a secret for a computer.",
    )

    def query(self, context: Account, call: Call) -> dict:
        account = context

        if not call.args.secrets_name:
            raise MissingParameterError("The secrets_name parameter must be specified")

        client_class = get_secrets_client()
        client = client_class(account.name, call.args.token)
        path = os.path.join(call.args.secrets_name, call.args.data_id)
        secret = {call.args.data_id: call.args.data.decode()}

        response = client.create_or_update_secret(path, secret)

        return response


@method
class GetSecret(AccountMethod):
    """Retrieves a secret."""

    requires = "ViewSecrets"
    route = AccountRoute()

    schema = Schema(
        name="GetSecret",
        parameters=[
            Unicode("data_id", doc="The name of the secret to be stored."),
            Integer(
                "computer_id",
                optional=True,
                doc="The numerical ID of the computer.",
            ),
            Unicode(
                "secrets_name",
                optional=True,
                doc="A unique identifier for the computer. Note that this is "
                "different from the computer_id.",
            ),
            Unicode(
                "token",
                optional=True,
                doc="The authentication token to be passed to the secrets " "manager.",
            ),
        ],
        errors=[
            InvalidParameterCombinationError,
            MissingParameterError,
            SecretNotFoundError,
            SecretsAuthenticationError,
            SecretsSealedError,
        ],
        doc="Retrieves a secret, one of either 'computer_id' or 'computer_name' "
        "must be provided.",
    )

    def query(self, context: Account, call: Call) -> dict:
        account = context

        if not call.args.computer_id and not call.args.secrets_name:
            raise MissingParameterError(
                "Request must contain either computer_id or secrets_name parameter"
            )

        if call.args.computer_id and call.args.secrets_name:
            raise InvalidParameterCombinationError(
                "Request cannot contain both computer_id and secrets_name parameters"
            )

        if call.args.computer_id:
            secrets_name = account.get_computer(call.args.computer_id).secrets_name
        else:
            secrets_name = call.args.secrets_name

        client_class = get_secrets_client()
        client = client_class(account.name, call.args.token)
        path = os.path.join(secrets_name, call.args.data_id)

        response = client.read_secret(path, call.args.data_id)
        return response


@method
class ListSecrets(AccountMethod):
    requires = "ViewSecrets"
    route = AccountRoute()

    schema = Schema(
        name="ListSecrets",
        parameters=[
            List(
                "computer_ids",
                Integer(),
                optional=True,
                doc="A comma separated list of numerical IDs to filter which "
                "computers will have their secrets listed (if the secret "
                "exists).",
            ),
            Unicode(
                "token",
                optional=True,
                doc="The authentication token to be passed to the secrets " "manager.",
            ),
        ],
        errors=[
            SecretNotFoundError,
            SecretsAuthenticationError,
            SecretsSealedError,
        ],
        doc="Lists all secrets accessible for this account / user.",
    )

    def query(self, context: Account, call: Call) -> list:
        account = context

        client_class = get_secrets_client()
        client = client_class(account.name, call.args.token)
        ids_and_names = self.get_ids_and_names(
            account=account,
            client=client,
            given_ids=call.args.computer_ids,
        )

        responses = []

        for computer_id, secrets_name in ids_and_names:
            response = client.list_secrets(secrets_name)
            responses.append(
                {
                    "secrets": response,
                    "secrets_name": secrets_name,
                    "computer_id": computer_id,
                }
            )

        return responses

    def get_ids_and_names(
        self,
        account: Account,
        client: SecretsClient,
        given_ids: Optional[list] = None,
    ) -> list:
        """
        Returns a list of (L{Computer.id}, L{Computer.secrets_name}) tuples.
        """
        ids_and_names = []

        if given_ids:
            computers = account.get_computers(computer_ids=given_ids)

            for computer in computers:
                secrets_name = None

                if computer.secrets_name:
                    secrets_name = computer.secrets_name.strip("/")

                ids_and_names.append((computer.id, secrets_name))
        else:
            # List secrets for all computers if not specified
            secrets_names = client.list_secrets()
            if not secrets_names:
                return ids_and_names
            computers = account.get_computers(secrets_names=secrets_names)
            # There are more secrets than there are computers
            # We need to match them up to find which secrets don't have a computer
            used_secrets = set()

            for computer in computers:
                ids_and_names.append((computer.id, computer.secrets_name.strip("/")))
                used_secrets.add(computer.secrets_name)

            for unused_secret in set(secrets_names) - used_secrets:
                ids_and_names.append((None, unused_secret))

        return ids_and_names


@method
class LinkSecretToComputer(AccountMethod):
    requires = "ViewAccount"
    route = AccountRoute()

    schema = Schema(
        name="LinkSecretToComputer",
        parameters=[
            Unicode("secrets_name", doc="Secrets name"),
            Integer("computer_id", doc="Computer Id"),
            Unicode("token", optional=True, doc="Token"),
        ],
        errors=[SecretNotFoundError, SecretsAuthenticationError],
        doc="Link secret to computer",
    )

    def query(self, context: Account, call: Call) -> dict:
        account = context

        client_class = get_secrets_client()
        client = client_class(account.name, call.args.token)
        secrets_names = client.list_secrets()

        if call.args.secrets_name not in secrets_names:
            raise SecretNotFoundError(
                f"The secrets_name {call.args.secrets_name} does not exist"
            )

        computer = account.get_computer(call.args.computer_id)
        computer.secrets_name = call.args.secrets_name

        return {}


@method
class UnlinkSecretToComputer(AccountMethod):
    requires = "ViewAccount"
    route = AccountRoute()

    schema = Schema(
        name="UnlinkSecretToComputer",
        parameters=[
            Unicode("secrets_name", doc="Secrets name"),
            Integer("computer_id", doc="Computer Id"),
        ],
        errors=[SecretNotFoundError],
        doc="Unlink secret to computer",
    )

    def query(self, context: Account, call: Call) -> dict:
        account = context

        computer = account.get_computer(call.args.computer_id)

        if computer.secrets_name != call.args.secrets_name:
            raise SecretNotFoundError(
                f"Computer with id {call.args.computer_id} is not linked to secrets name {call.args.secrets_name}"
            )

        computer.secrets_name = None

        return {}
