# Copyright 2005-2023 Canonical Limited. All rights reserved.
"""
Secrets Client
==============

The classes here are used to interface with external secrets managers.
"""

import abc
import base64
import os

from pkg_resources import get_distribution, parse_version
from typing import Optional

import requests

from zope.component import getUtility

from canonical.landscape.interfaces import IFileBasedConfiguration
from canonical.landscape.api.error import (
    SecretsAuthenticationError,
    SecretsSealedError,
    SecretNotFoundError,
)


DEFAULT_VAULT_URL = "http://localhost:8200"


class SecretsClient(abc.ABC):
    """The interface for secrets management engines."""

    def __init__(
        self, account_name: str, token: Optional[str], url: Optional[str] = None
    ):
        self.account_name = account_name
        if token:
            self.token = token
        else:
            self.token = get_token()
        if url:
            self.url = url
        else:
            self.url = get_vault_url()

    @abc.abstractmethod
    def create_or_update_secret(self):
        ...

    @abc.abstractmethod
    def read_secret(self):
        ...

    @abc.abstractmethod
    def list_secrets(self):
        ...


def get_token():
    """Gets the token from either environmental variable, or secrets service. In that order."""
    token_env = os.getenv("VAULT_TOKEN")
    if token_env:
        return token_env
    config = getUtility(IFileBasedConfiguration)
    url = config.get("secrets", "secrets-service-url", "")
    error_msg = (
        "No Vault token specified and unable to reach Landscape secrets service!"
    )
    if not url:
        raise SecretsAuthenticationError(error_msg)
    try:
        token = requests.get(url).json()["token"]
    except requests.exceptions.ConnectionError:
        raise SecretsAuthenticationError(error_msg)
    return token


def get_vault_url():
    """Gets the vault address from either, environmental variable, config file, or default value In that order."""
    addr_env = os.getenv("VAULT_ADDR")
    if addr_env:
        return addr_env
    config = getUtility(IFileBasedConfiguration)
    addr_conf = config.get("secrets", "secrets-url", "")
    if addr_conf:
        return addr_conf
    else:
        return DEFAULT_VAULT_URL


class HTTPVaultClient(SecretsClient):
    """Interface for Hashicorp Vault via HTTP.

    Reference:
        https://developer.hashicorp.com/vault/api-docs
    """

    def __init__(
        self, account_name: str, token: Optional[str] = None, url: Optional[str] = None
    ):
        super().__init__(account_name, token, url)

        self.base_url = os.path.join(self.url, "v1")
        self.account_url = os.path.join(self.base_url, self.account_name)
        self.headers = {"X-Vault-Token": self.token}

        self._create_account()

    def create_or_update_secret(self, path: str, secret: dict) -> dict:
        url = os.path.join(self.account_url, "data", path)
        json_data = {"data": secret}

        response = self._request("POST", url=url, headers=self.headers, json=json_data)
        created_time = response.get("data", {}).get("created_time")

        return {"created_time": created_time}

    def read_secret(self, path: str, secret_name: str) -> dict:
        url = os.path.join(self.account_url, "data", path)

        response = self._request("GET", url=url, headers=self.headers)

        return response.get("data", {})

    def delete_secret(self, path: str) -> None:
        url = os.path.join(self.account_url, "metadata", path)

        response = self._request("DELETE", url=url, headers=self.headers)

        return response

    def list_secrets(self, path: str = "") -> list:
        url = os.path.join(self.account_url, "metadata", path)

        try:
            response = self._request("LIST", url=url, headers=self.headers)
        except SecretNotFoundError:
            return []  # Will raise this if no secrets
        secrets = response.get("data", {}).get("keys", [])
        secrets = [secret.strip("/") for secret in secrets]

        return secrets

    def create_orphan_token(self, ttl: str, policy_name: str) -> str:
        url = os.path.join(self.base_url, "auth/token/create-orphan")
        json_data = {"ttl": ttl, "policies": [policy_name]}
        response = self._request("POST", url=url, headers=self.headers, json=json_data)
        return response["auth"]["client_token"]

    def revoke_token(self, token) -> dict:
        url = os.path.join(self.base_url, "auth/token/revoke")
        json_data = {"token": token}
        response = self._request("POST", url=url, headers=self.headers, json=json_data)
        return response

    def get_policy(self, name: str) -> dict:
        url = os.path.join(self.base_url, "sys/policies/acl", name)
        response = self._request("GET", url=url, headers=self.headers)
        return response

    def create_policy(self, name: str, policy: str) -> dict:
        url = os.path.join(self.base_url, "sys/policies/acl", name)
        policy = base64.b64encode(policy.encode()).decode()
        json_data = {"policy": policy}
        response = self._request("POST", url=url, headers=self.headers, json=json_data)
        return response

    def _create_account(self) -> None:
        """Check if the account exists, otherwise create a new mount."""
        url = os.path.join(self.base_url, "sys/mounts", self.account_name, "tune")
        try:
            self._request("GET", url=url, headers=self.headers)
        except SecretNotFoundError:
            url = os.path.join(self.base_url, "sys/mounts", self.account_name)
            json_data = {"type": "kv-v2"}
            self._request("POST", url=url, headers=self.headers, json=json_data)

    def _request(self, *args, **kwargs):
        try:
            response = requests.request(*args, **kwargs)
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.InvalidSchema,
            requests.exceptions.MissingSchema,
            requests.exceptions.Timeout,
        ) as error:
            raise SecretsAuthenticationError(str(error))

        if not response.ok:
            self._raise_error(response.status_code)

        if not response.text:  # Sometimes response is empty therefore not json
            return

        return response.json()

    @staticmethod
    def _raise_error(status_code: int):
        if status_code == 403:
            raise SecretsAuthenticationError(
                "Vault authentication failed, please double check your token"
            )
        elif status_code == 404:
            raise SecretNotFoundError("Secret could not be found")
        elif status_code == 503:
            raise SecretsSealedError("Vault is sealed")
        else:
            raise SecretNotFoundError("Vault could not be reached")


def get_secrets_client():
    return HTTPVaultClient


def get_secret(account_name, token, secrets_name, data_id):
    secrets_client_class = get_secrets_client()
    client = secrets_client_class(account_name, token)
    path = os.path.join(secrets_name, data_id)
    secret = client.read_secret(path, data_id)
    return secret
