from collections import defaultdict
from typing import Dict, List

from canonical.landscape.api import RESULT_OFFSET, RESULT_LIMIT
from canonical.landscape.api.helpers import (
    paginated,
    with_computer_selection,
    with_computer,
    ApiRequestError,
)
from canonical.landscape.api.schema import Array
from canonical.landscape.model.package.management import (
    PackageManagement,
    ComputerChangeTuple,
)
from canonical.landscape.model.package.usn import parse_usn
from canonical.landscape.model.package.search import (
    PackageSearch,
    fill_usn_counts,
    get_num_usns_by_computers,
    get_usn_query,
)
from canonical.landscape.model.resource.store import get_resource_stores
from canonical.txapi.server.schema import (
    Bool,
    Integer,
    PositiveInteger,
    Schema,
    Unicode,
)


USN_SCHEMA = Schema(
    name="Usns",
    parameters=[
        Array(
            "computer_ids",
            Integer(),
            optional=False,
            doc="Get USN information for these computers",
        ),
        Bool(
            "show_packages",
            default=False,
            optional=True,
            doc=("Whether or not to show packages associated with usns"),
        ),
        PositiveInteger(
            "limit",
            default=RESULT_LIMIT,
            optional=True,
            doc=(
                "The maximum number of results returned by the "
                "method. It defaults to 1000."
            ),
        ),
        Integer(
            "offset",
            default=RESULT_OFFSET,
            optional=True,
            doc="The offset inside the list of results.",
        ),
    ],
    doc="""Get security upgrades info associated with these computers grouped by usns (includes cves)""",
)


def get_usns(usn_counts):
    """
    Gets the usn information for the selected computers formatted for the api
    """

    usn_query = get_usn_query(usn_counts.keys())

    results = []
    for (
        usn,
        cves,
    ) in usn_query:
        usn_name = usn.name
        date = usn.release_time.isoformat()
        packages = usn.summary
        cve_data = []
        for cve in cves:
            cve_data.append(
                {
                    "cve": cve,
                    "cve_link": "https://cve.mitre.org/cgi-bin/cvename.cgi?name=" + cve,
                }
            )
        data = {
            "usn": usn.name,
            "usn_link": "https://ubuntu.com/security/notices/USN-" + usn.name,
            "date": usn.release_time.isoformat(),
            "info": usn.summary,
            "cves": cve_data,
            "computers_count": usn_counts[usn.id],
        }
        results.append(data)

    return results


def get_packages_by_usns(package_search, usn_name=None):
    """
    Takes the data from get_relations_by_package_name and reformats it for the api
    such that it is grouped by usn_name
    """

    result = package_search.get_relations_by_package_name(
        upgrades=True, limit=None, usn=True, usn_name=usn_name
    )
    packages = result["packages"]
    usns_to_packages = defaultdict(list)
    for relation in result["relations"]:
        for rel_info in relation["relations"]:
            computer_ids = rel_info["computers"]
            installed_id = rel_info["installed"]
            installed = packages[installed_id]
            upgrade_id = rel_info["upgrade"]
            upgrade = packages[upgrade_id]
            usn = upgrade.usn
            if usn:
                package_info = {
                    "name": upgrade.name,
                    "current_id": installed.id,
                    "current_version": installed.version,
                    "new_version": upgrade.version,
                    "summary": upgrade.summary,
                    "computer_ids": computer_ids,
                }
                usns_to_packages[usn.name].append(package_info)
    if usn_name:  # Since it's just one then return just this item
        return usns_to_packages[usn_name]
    else:
        return usns_to_packages


@with_computer_selection(USN_SCHEMA, "ViewComputer")
@paginated
async def usns_list_handler(
    request,
    principal,
    selection,
    validated_data,
    limit=RESULT_LIMIT,
    offset=RESULT_OFFSET,
):

    ids = validated_data["computer_ids"]
    show_packages = validated_data["show_packages"]

    package_search = None
    usns_to_packages = {}
    if show_packages:
        package_search = PackageSearch(selection)
        usns_to_packages = get_packages_by_usns(package_search)

    # Note this endpoint does not currently support sharding
    store = list(get_resource_stores())[0][1]
    usn_counts = defaultdict(int)
    fill_usn_counts(usn_counts, store, ids, limit=limit, offset=offset)

    usns = get_usns(usn_counts)
    for usn in usns:
        usn_name = usn["usn"]
        if show_packages:
            usn["packages"] = usns_to_packages.get(usn_name, [])

    return {
        "count": get_num_usns_by_computers(store, ids),
        "results": usns,
    }, 200


@with_computer_selection(USN_SCHEMA, "ViewComputer")
async def usns_get_handler(
    request,
    principal,
    selection,
    validated_data,
    url_args,
    limit=RESULT_LIMIT,
    offset=RESULT_OFFSET,
):

    ids = validated_data["computer_ids"]
    usn_name = url_args["usn_name"]
    try:
        parse_usn(usn_name)
    except ValueError as exc:
        raise ApiRequestError(message=f"Usn name '{usn_name}' is invalid!") from exc

    package_search = PackageSearch(selection)
    result = get_packages_by_usns(package_search, usn_name=usn_name)

    return result, 200


USN_ACTIVITY_SCHEMA = Schema(
    name="Usns",
    parameters=[
        Array(
            "usns",
            Unicode(),
            optional=False,
            doc="List of USN names for the command",
        ),
    ],
    doc="Create an activity to upgrade or remove packages affected by selected USNs",
)


def get_packages_matching_usns(
    package_search: PackageSearch, usns: List[str]
) -> List[Dict]:
    """
    Returns a list of packages matching the USN names provided
    """
    result = get_packages_by_usns(package_search)
    packages_matched = []
    for usn_name in usns:
        packages = result[usn_name]
        packages_matched.extend(packages)
    return packages_matched


class NoUsnPackagesMatched(ApiRequestError):
    def __init__(self, usns):
        self.message = f"No packages for this computer are associated with USNs {usns}"


@with_computer(USN_ACTIVITY_SCHEMA, "ViewComputer")
async def usns_upgrade_handler(
    request,
    principal,
    computer,
    selection,
    validated_data,
    url_args,
    limit=RESULT_LIMIT,
    offset=RESULT_OFFSET,
):

    usns = validated_data["usns"]

    management = PackageManagement(selection)
    package_search = management.get_search()

    packages = get_packages_matching_usns(package_search, usns)
    package_names = list(set([p["name"] for p in packages]))

    if not package_names:
        raise NoUsnPackagesMatched(usns)

    changes = package_search.get_package_upgrades_by_computer_id(names=package_names)

    activity = management.change_packages(
        principal.person,
        changes,
    )
    return activity.get_state(), 202


@with_computer(USN_ACTIVITY_SCHEMA, "ViewComputer")
async def usns_remove_handler(
    request,
    principal,
    computer,
    selection,
    validated_data,
    url_args,
    limit=RESULT_LIMIT,
    offset=RESULT_OFFSET,
):

    usns = validated_data["usns"]

    management = PackageManagement(selection)
    package_search = management.get_search()

    packages = get_packages_matching_usns(package_search, usns)
    package_ids = list(set([p["current_id"] for p in packages]))

    if not package_ids:
        raise NoUsnPackagesMatched(usns)

    change = ComputerChangeTuple(
        computer=url_args["computer_id"],
        installs=(),
        removals=package_ids,
        holds=(),
        hold_removals=(),
    )

    activity = management.change_packages(
        principal.person,
        [change],
    )
    return activity.get_state(), 202
