import asyncio
import datetime
import json
import logging
import subprocess

import aio_pika
from storm.locals import Store
from storm.zope.interfaces import ZStormError
import transaction

from canonical.landscape.application import setup_logging
from canonical.landscape.hostagent.config import AMQP_CONFIG
from canonical.landscape.message.handlers.ubuntu_pro import update_pro_licensing
from canonical.landscape.model.computer.management import ComputerManagement
from canonical.landscape.model.main.account import get_account_by_name
from canonical.landscape.model.main.computer import (
    cleanup_after_removed_computers,
    get_computer_by_secure_id,
    get_pending_computer_by_secure_id,
)
from canonical.landscape.model.main.distribution import (
    get_distribution,
    create_distribution,
    WSL_HOST_CODE_NAME,
)
from canonical.landscape.model.main.store import get_main_store
from canonical.landscape.model.selection.selection import ComputerSelection
from canonical.landscape.setup import load_config


class HostagentMessageConsumer:
    def __init__(self):
        self.host = AMQP_CONFIG["host"]
        self.port = AMQP_CONFIG["port"]
        self.login = AMQP_CONFIG["user"]
        self.password = AMQP_CONFIG["password"]
        self.virtualhost = AMQP_CONFIG["vhost"]
        self.queue_name = AMQP_CONFIG["hostagent_task_queue"]

    async def consume(self):
        connection = await aio_pika.connect_robust(
            host=self.host,
            port=self.port,
            login=self.login,
            password=self.password,
            virtualhost=self.virtualhost,
        )

        async with connection:
            channel = await connection.channel()

            await channel.set_qos(prefetch_count=10)
            queue = await channel.declare_queue(self.queue_name)

            async with queue.iterator() as queue_iter:
                async for message in queue_iter:
                    async with message.process():
                        await self.handle_message(message.body.decode())

    async def handle_message(self, message):
        parsed_message = json.loads(message)

        hostagent_uid = parsed_message.get("uid")
        pro_token = parsed_message.get("token")
        hostname = parsed_message.get("hostname")
        instances = parsed_message.get("instances")
        account_name = parsed_message.get("account_name")
        registration_password = parsed_message.get("registration_key")
        default_instance_id = parsed_message.get("default_instance_id")

        account = None
        if account_name:
            account = get_account_by_name(account_name)

        if not account:
            logging.info(
                "Account %s not found when processing %s", account_name, hostname
            )
            return

        computer = get_computer_by_secure_id(hostagent_uid)
        if computer:
            await self.process_computer(
                computer, pro_token, hostname, default_instance_id
            )
            await self.process_instances(computer, instances)
            return
        else:
            logging.info("computer %s not found for account %s", hostname, account_name)

        pending_computer = get_pending_computer_by_secure_id(hostagent_uid)

        if not pending_computer:
            valid_registration_password = account.check_registration_password(
                registration_password
            )
            if not valid_registration_password:
                logging.info(
                    "Invalid registration key provided by %s for account %s",
                    hostname,
                    account_name,
                )
                return
            logging.info(
                "Creating pending computer %s for account %s", hostname, account_name
            )
            pending_computer = await self.create_pending_computer(
                account, hostagent_uid, pro_token, hostname
            )
        else:
            logging.info(
                "Pending computer %s found for account %s.", hostname, account_name
            )
            logging.info(pending_computer.id)

        if account.may_auto_register_computers:
            computer = pending_computer.accept(person=None)
            logging.info("Accepting computer %s to account %s.", hostname, account_name)
            computer.store.commit()
            computer.account.store.commit()

        # We call commit on the store to clear pending_computer from the cache
        get_main_store().commit()

    async def process_computer(
        self, computer, pro_token, hostname, default_instance_id=None
    ):
        with transaction.manager as manager:
            self.update_ping_time(computer)
            computer.default_child_instance = default_instance_id

            if not computer.ubuntu_pro_info:
                await self.assign_ubuntu_pro_info(computer, pro_token)

            computer.hostname = hostname

    async def process_instances(self, computer, instances):
        computer_children = computer.get_hostagent_child_computers()
        instance_names = [instance["id"] for instance in instances]
        children_to_remove = []
        for computer_child in computer_children:
            if computer_child.title not in instance_names:
                children_to_remove.append(computer_child.id)
        if children_to_remove:
            await self.remove_children(
                computer.account,
                children_to_remove,
            )
            logging.info(
                f"Removed these instances from account {computer.account.name}: {children_to_remove}"
            )

    async def remove_children(self, account, computer_ids):
        with transaction.manager as manager:
            selection = ComputerSelection(account)
            selection.select_computers(computer_ids)
            management = ComputerManagement(selection, None)
            management.remove_computers()
            cleanup_after_removed_computers(account, computer_ids)

    async def create_pending_computer(self, account, secure_id, pro_token, hostname):
        ubuntu_pro_info = await self.get_ubuntu_pro_info(pro_token)
        pending_computer = None
        with transaction.manager as manager:
            distribution = get_distribution(code_name=WSL_HOST_CODE_NAME)
            if not distribution:
                logging.info(
                    "Creating distribution with codename %s", WSL_HOST_CODE_NAME
                )
                distribution = create_distribution(
                    "Windows 10 / Windows 11",
                    "Microsoft",
                    "10 / 11",
                    WSL_HOST_CODE_NAME,
                )
            try:
                pending_computer = account.create_pending_computer(
                    title=hostname,
                    description=f"Host machine {hostname}",
                    hostname=hostname,
                    ubuntu_pro_info=ubuntu_pro_info,
                    distribution_id=distribution.id,
                )
                pending_computer.secure_id = secure_id
            except Exception:
                logging.exception("Error creating pending computer.")
                manager.abort()
        return pending_computer

    def update_ping_time(self, computer):
        computer.last_ping_time = datetime.datetime.utcnow()

    async def get_ubuntu_pro_info(self, pro_token):
        # This process could take some time to complete.
        try:
            completed_process = await asyncio.create_subprocess_exec(
                "ua",
                "status",
                "--simulate-with-token",
                pro_token,
                "--format",
                "json",
                stdout=subprocess.PIPE,
            )
        except FileNotFoundError:
            # ubuntu-advantage-tools is not available
            return None
        else:
            ubuntu_pro_info = await completed_process.stdout.read()
            return ubuntu_pro_info.decode().strip()

    async def assign_ubuntu_pro_info(self, computer, pro_token):
        ubuntu_pro_info = await self.get_ubuntu_pro_info(pro_token)
        if ubuntu_pro_info:
            logging.info(f"updating ubuntu pro info for computer id {computer.id}")
            computer.ubuntu_pro_info = ubuntu_pro_info
            update_pro_licensing(computer)


async def consume():
    setup_logging("hostagent-message-consumer")
    load_config("hostagent-message-consumer")
    consumer = HostagentMessageConsumer()
    await consumer.consume()
