# Copyright 2005-2022 Canonical Limited. All rights reserved.

import os
from email import charset, encoders
from email.header import Header
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.image import MIMEImage
from email.utils import formataddr, parseaddr

from zope.interface import implementer
from zope.component import getUtility
from zope.sendmail.interfaces import IMailDelivery

from canonical.lib.mail import idna_encode_email_address
from canonical.landscape.mail.interfaces import ILandscapeMailer
from canonical.landscape.model.configuration import (
    ISystemConfiguration,
    LANDSCAPE_SYSTEM_EMAIL,
)


# By default the content transfer encoding for utf-8 is base64. We make
# it quoted printable, so that the e-mails can be read without having to
# decode them first.
del charset.CHARSETS["utf-8"]
charset.add_charset("utf-8", charset.SHORTEST, charset.QP, "utf-8")


def format_address(name, email):
    """Format a name and email as an e-mail header string.

    It works like email.Util.formataddr(), except that it handles
    unicode name properly.
    """
    email = idna_encode_email_address(email)
    if not name:
        return email
    name = str(Header(name))
    return formataddr((name, email))


@implementer(ILandscapeMailer)
class LandscapeMailer:
    """A convenience wrapper around zope.sendmail.delivery.QueuedMailDelivery.

    This class provides a more friendly interface for sending mails from
    the application code.
    """

    def send(
        self,
        to_addresses,
        subject,
        message_content,
        sender=None,
        bulk=False,
        attachments=None,
    ):
        """Send an email:

        @param to_addresses: a list of addresses to sent the email to
        @param subject: the subject of the email
        @param message_content: the content of the email
        @param sender: optionally the email sender
        @param bulk: if True, set the Precedence header to Bulk
        @param attachments: a 2-tuple of (filename, content-type) of files
            to attach to the email
        """
        mailer = getUtility(IMailDelivery, "landscape-mailer")
        text_message = MIMEText(message_content.encode("utf-8"), "plain", "utf-8")
        if attachments:
            message = MIMEMultipart()
            message.attach(text_message)
        else:
            message = text_message

        configuration = getUtility(ISystemConfiguration)
        if sender is None:
            sender = configuration.get(
                LANDSCAPE_SYSTEM_EMAIL, "landscape-team@canonical.com"
            )
        message["From"] = sender
        message["To"] = ", ".join(to_addresses)
        message["Subject"] = subject
        if bulk:
            message.add_header("Precedence", "Bulk")
        if attachments:
            add_attachments(message, attachments)
        mail_addresses = [parseaddr(address)[1] for address in to_addresses]
        mailer.send(sender, mail_addresses, message.as_string())


@implementer(ILandscapeMailer)
class LandscapeHTMLMailer:
    """A convenience wrapper around zope.sendmail.delivery.QueuedMailDelivery.

    This class provides a more friendly interface for sending mails from
    the application code.
    """

    def send(
        self,
        to_addresses,
        subject,
        plaintext_content,
        html_content=None,
        sender=None,
        bulk=False,
        attachments=None,
        inline_media=None,
    ):
        """Send an HTML-enabled email:

        :param list[str] to_addresses: a list of addresses to sent the email to
        :param str subject: the subject of the email
        :param str plaintext_content: the plaintext version of the email content.
        :param str html_content: the HTML version of the email content.
        :param str sender: optionally the email sender
        :param bool bulk: if True, set the Precedence header to Bulk
        :param list attachments: a list of 2-tuples of (filename, content-type) of
            files to attach to the email
        :param list inline_media: a list of 3-tuples with format
            (filename, content-type, content-id) corresponding to files to include
            in the main body of the HTML version of the email. Must be identified
            in the HTML body by the corresponding Content-ID.
        """
        mailer = getUtility(IMailDelivery, "landscape-mailer")
        message = self._build_message(
            plaintext_content, html_content, attachments, inline_media
        )

        configuration = getUtility(ISystemConfiguration)
        if sender is None:
            sender = configuration.get(
                LANDSCAPE_SYSTEM_EMAIL, "landscape-team@canonical.com"
            )
        message["From"] = sender
        message["To"] = ", ".join(to_addresses)
        message["Subject"] = subject
        if bulk:
            message.add_header("Precedence", "Bulk")
        mail_addresses = [parseaddr(address)[1] for address in to_addresses]
        mailer.send(sender, mail_addresses, message.as_string())

    def _build_message(self, plaintext, html, attachments, inline_media):
        """Construct an appropriately-shaped MIME message given the intended content"""
        plaintext = MIMEText(plaintext.encode("utf-8"), "plain", "utf-8")
        if html is None and attachments is None:
            return plaintext
        elif html is None and attachments:
            message = MIMEMultipart()
            message.attach(plaintext)
            add_attachments(message, attachments)
            return message
        elif html and attachments is None:
            message = MIMEMultipart("alternative")
            message.attach(plaintext)
            html_message = self._build_html_message(html, inline_media)
            message.attach(html_message)
            return message
        elif html and attachments:
            message = MIMEMultipart("mixed")
            message_body = MIMEMultipart("alternative")
            message_body.attach(plaintext)
            html_message = self._build_html_message(html, inline_media)
            message_body.attach(html_message)
            message.attach(message_body)
            add_attachments(message, attachments)
            return message

    def _build_html_message(self, html_content, inline_media):
        """Construct an HTML message that may include inline media"""
        if inline_media:
            html_message = MIMEMultipart("related")
            html_message.attach(MIMEText(html_content.encode("utf-8"), "html", "utf-8"))
            self._add_inline_media(html_message, inline_media)
        else:
            html_message = MIMEText(html_content.encode("utf-8"), "html", "utf-8")
        return html_message

    def _add_inline_media(self, message, media):
        """Add inline content to the message.  Should be identified in HTML
        body by Content-ID.
        """
        for filename, content_type, content_id in media:
            part = MIMEBase(*content_type.split("/", 1))
            with open(filename, "rb") as fd:
                part.set_payload(fd.read())
            encoders.encode_base64(part)
            part.add_header(
                "Content-Disposition", "inline", filename=os.path.basename(filename)
            )
            part.add_header("Content-ID", content_id)
            message.attach(part)


def add_attachments(message, attachments):
    """Attach files to the message.

    @param message the MIME object to attach to
    @param attachments a 2-tuple of (filename, content-type) of files
        to attach to the email
    """
    for attachment, content_type in attachments:
        part = MIMEBase(*content_type.split("/", 1))
        with open(attachment, "rb") as fd:
            part.set_payload(fd.read())
        encoders.encode_base64(part)
        # Set the filename parameter
        part.add_header(
            "Content-Disposition",
            "attachment",
            filename=os.path.basename(attachment),
        )
        message.attach(part)


landscape_mailer = LandscapeHTMLMailer()
