# Copyright 2005-2022 Canonical Limited. All rights reserved.

"""Functionality for managing account details."""

from storm.locals import Store, Count

from canonical.security.protect import requires, requires_advisor

from canonical.landscape.model.management import ModelManagement
from canonical.landscape.model.main.personaccess import (
    get_access_context_ids_for_person,
)
from canonical.landscape.ui.salesforce.key import validate_salesforce_account_key
from canonical.landscape.ui.format import is_valid_registration_key


__all__ = ["AccountManagement"]

__clearlist__ = ["AccountManagement"]

__requires_prefix__ = "landscape."


@requires_advisor
class AccountManagement(ModelManagement):
    """An API for getting information about an L{Account}."""

    @requires("ViewAccount")
    def get_access_contexts_with_counts(self):
        """Get account L{AccessContext}s with a count of L{Computer}s for each.

        @return: A list of of C{dict}s with keys C{access_context} and
            C{computers_count}, sorted by access context name.
        """
        from canonical.landscape.model.main.accesscontext import AccessContext
        from canonical.landscape.model.main.computer import Computer

        account = self.account
        if self.person.has_permission("landscape.ViewComputer", account.access_context):
            # Person has account-wide permissions, grab all contexts.
            access_context_ids = [
                context.id for context in account.get_access_contexts()
            ]
        else:
            # Person is restricted admin, check specific access contexts.
            access_context_ids = get_access_context_ids_for_person(
                account.access_context, "landscape.ViewComputer", self.person
            )
        result = Store.of(account).find(
            (AccessContext, Count(Computer.id)),
            Computer.access_context_id == AccessContext.id,
            AccessContext.id.is_in(access_context_ids),
        )

        result.group_by(AccessContext)
        result.having(Count(Computer.id) > 0)
        result.order_by(AccessContext.name)
        return [
            {"access_context": access_context, "computers_count": count}
            for access_context, count in result
        ]

    @requires("ViewAccount")
    def get_tags_with_counts(self):
        """Get account L{Tag}s with a count of L{Computer}s for each one.

        @return: A list of of C{dict}s with keys C{tag} and C{computers_count},
            sorted by tag name.
        """
        from canonical.landscape.model.main.computer import Computer
        from canonical.landscape.model.main.tag import Tag, ComputerTag

        account = self.account
        where = [Tag.id == ComputerTag.tag_id]
        computer_ids = self.person.get_computer_ids(account)
        if computer_ids is None:
            where.extend(
                [
                    ComputerTag.computer_id == Computer.id,
                    Computer.account_id == account.id,
                ]
            )
        else:
            where.append(ComputerTag.computer_id.is_in(computer_ids))
        result = Store.of(account).find((Tag, Count(ComputerTag.computer_id)), *where)
        result.group_by(Tag)
        result.having(Count(ComputerTag.computer_id) > 0)
        result.order_by(Tag.name)
        return [{"tag": tag, "computers_count": count} for tag, count in result]

    def disable_membership(self, person):
        """
        Disable a person's membership in the account and log the event.
        """
        self.account.disable_membership(person)
        self._event_log.create_log_entry(
            self.account,
            "Account access disabled for %s (%s)." % (person.name, person.email),
        )

    @requires("ViewAccount")
    def get_preferences(self):
        preferences = {
            "title": self.account.title,
            "registration_password": self.account.registration_password,
            "auto_register_new_computers": self.account.auto_register_new_computers,
        }

        return preferences

    @requires("EditAccount")
    def set_preferences(
        self,
        title=None,
        registration_password=None,
        auto_register_new_computers=None,
    ):
        if registration_password is not None:
            if is_valid_registration_key(registration_password):
                self.account.registration_password = registration_password
            else:
                raise ValueError(
                    f"Invalid registration key format '{registration_password}'"
                )
        if title is not None:
            self.account.title = title
        if auto_register_new_computers is not None:
            self.account.auto_register_new_computers = auto_register_new_computers

    @requires("EditSalesforceKey")
    def set_salesforce_account_key(self, salesforce_account_key):
        validate_salesforce_account_key(salesforce_account_key)
        self.account.salesforce_account_key = salesforce_account_key
