# Copyright 2005-2016 Canonical Limited.  All rights reserved.
import json
import logging

from storm.expr import Desc
from twisted.python.failure import Failure
from zope.component import getUtility

from canonical.amqp.message.interfaces import IDispatcher
from canonical.landscape.hostagent.config import AMQP_CONFIG
from canonical.landscape.model.account.store import get_account_store_by_name
from canonical.landscape.model.main.account import get_account_by_id
from canonical.landscape.model.activity.activity import (
    ActivityInfo,
    AbstractActivityInfo,
    Activity,
    get_child_activity,
)
from canonical.landscape.model.activity.computer import ComputerRequestMixin
from canonical.landscape.model.activity.parent import ActivityParent
from canonical.landscape.model.computer.twin import AccountComputerTwin
from canonical.landscape.model.activity.status import BLOCKED
from canonical.landscape.model.activity.collection import (
    ActivityParentCollection,
    ActivityChildCollection,
)
from canonical.landscape.model.eventlog.management import EventLog
from canonical.security.protect import requires, requires_advisor


__all__ = [
    "ChildComputerKeepAliveMixin",
    "CreateChildComputerActivity",
    "DeleteChildComputerActivity",
    "get_computer_activities_parents",
    "get_next_parent_with_blocked_computer_activity",
    "GRPCActivityMixin",
    "SetDefaultChildComputerActivity",
    "ShutdownHostComputerActivity",
    "StartChildComputerActivity",
    "StopChildComputerActivity",
]

__clearlist__ = __all__


def get_computer_activities_parents(
    account, computer_ids=None, statuses=None, info_classes=None, ordered=False
):
    """Return parent activities with children related to specified computers.

    @param account: the Account to search in.
    @param computer_ids: optionally IDs of the computers to search activities
        for. If no IDs are specified, activities for all computers are
        searched.
    @param statuses: optionally limit child activities to specified statuses.
    @param info_classes: optionally limit child activities to the specified
        info classes.
    @param ordered: whether to order parents by creation time and id.
    """
    if not info_classes:
        # Allow empty sequences as parameter
        info_classes = None

    child_collection = ActivityChildCollection(
        account.store,
        account=account,
        statuses=statuses,
        info_classes=info_classes,
        entity_class=AccountComputerTwin,
        entity_ids=computer_ids,
    )
    result = child_collection.find()
    result = result.config(distinct=True)
    parent_ids = result.values(Activity.parent_id)

    # Return parents of the children previously found
    parent_collection = ActivityParentCollection(
        account.store, account=account, ids=parent_ids
    )
    result = parent_collection.find()
    if ordered:
        result = result.order_by(
            Desc(ActivityParent.creation_time), Desc(ActivityParent.id)
        )
    return result.find()


def get_next_parent_with_blocked_computer_activity(activity_parent, computer_ids=None):
    """Return the next parent activity with children in BLOCKED status.

    Parent activities are considered based on their creation time.

    If computer_ids are passed, only parents which children involving specified
    computers are considered.
    """
    account = activity_parent.account

    parents_result = get_computer_activities_parents(
        account, computer_ids=computer_ids, statuses=[BLOCKED]
    )
    parents_result.order_by(ActivityParent.creation_time)
    result = parents_result.find(
        ActivityParent.creation_time > activity_parent.creation_time
    )
    next_blocked = result.first()

    if next_blocked is None:
        # It could be that we got the last activity, try to loop around to the
        # first blocked activity.
        result = parents_result.find(
            ActivityParent.creation_time < activity_parent.creation_time
        )
        next_blocked = result.first()

    return next_blocked


@requires_advisor
class GRPCActivityMixin(ComputerRequestMixin):
    """
    Base class for Activities that should put a message on the queue
    for the GRPC server to send along to hostagents.

    Subclasses should implement the `queue`, `message`, `get_summary`,
    `should_cancel_when_queued`, and `cancel_message` methods. In addition,
    they should override the `_data` method to serialize any data required
    by the activity.

    Subclasses should also inherit from `AbstractActivityInfo` if they
    require data storage. Otherwise they should also inherit from
    `ActivityInfo`.
    """

    # Always `False` so we don't spawn activities forever.
    must_keep_instance_alive = False

    def __init__(self, activity, computer, child_computer=None):
        self.activity = activity
        self.data = self._data(child_computer)
        if computer:
            self._set_activity_entity(computer.get_account_twin())
        else:
            self._set_activity_entity(child_computer.get_account_twin())
        self.activity_id = activity.id
        self.account_id = activity.account_id
        self.queue = self.computer.secure_id.decode() if computer else ""
        self.message = self._message

    @property
    def _message(self):
        """
        Subclasses should construct the message to put on the queue
        for the GRPC message relay server to send to the hostagent.
        """
        raise NotImplementedError()

    def get_summary(self):
        """
        Summary description of the activity.
        """
        raise NotImplementedError()

    @property
    def should_cancel_when_queued(self):
        """
        If `True` we cancel the activity immediately when queued.
        If `False` we start the activity.
        """
        raise NotImplementedError()

    @property
    def cancel_message(self):
        """
        Message to display when the activity is
        automatically canceled.
        """
        raise NotImplementedError()

    def _data(self, child_computer):
        """
        Subclasses should serialize any required data.
        """
        return {}

    def _queued(self):
        if self.should_cancel_when_queued:
            self.activity.cancel(text=self.cancel_message)
        else:
            self.activity.start()

    def _started(self):
        try:
            success = self.send_grpc_message()
        except Exception as e:
            self._fail(Failure(e))
        else:
            self._succeed_activity(success)

    def _succeed_activity(self, success):
        self.activity.succeed(text="Message queued.")

    def _fail(self, failure):
        """Mark the activity as failed."""
        logging.error(failure)
        account = get_account_by_id(self.account_id)
        activity = get_child_activity(account, self.activity_id)
        activity.fail(text=str(failure.value))

    def send_grpc_message(self):
        """Queue the GRPC message to be sent on transaction commit"""

        dispatcher = getUtility(IDispatcher)
        dispatcher.send_grpc_message(self.queue, self.message)


@requires_advisor
class CreateChildComputerActivity(GRPCActivityMixin, AbstractActivityInfo):
    """
    Activity to put a message on the queue for a hostagent to create a child
    computer instance.
    """

    def __init__(self, activity, computer, instance_name, cloud_init=None):
        self.activity = activity
        self.data = self._data(instance_name)
        self._set_activity_entity(computer.get_account_twin())
        self.activity_id = activity.id
        self.account_id = activity.account_id
        self.queue = self.computer.secure_id.decode()
        self.cloud_init = cloud_init
        self.message = self._message

    @requires("ViewActivity")
    def get_summary(self):
        return f"Create instance {self.data['instance_name']}"

    @property
    def _message(self):
        return json.dumps(
            {
                "install": {
                    "id": self.data.get("instance_name"),
                    "cloud_init": self.cloud_init,
                }
            }
        )

    @property
    def should_cancel_when_queued(self):
        return not self.computer.create_instance_supported

    @property
    def cancel_message(self):
        return f"Create instance not supported for computer {self.computer.title}"

    def _data(self, instance_name):
        return {
            "instance_name": instance_name,
        }


@requires_advisor
class DeleteChildComputerActivity(GRPCActivityMixin, AbstractActivityInfo):
    """
    Activity to put a message on the queue for a hostagent to delete a child
    computer instance.
    """

    @requires("ViewActivity")
    def get_summary(self):
        return f"Delete instance {self.data.get('title')}"

    @property
    def _message(self):
        return json.dumps({"uninstall": self.data.get("title")})

    @property
    def should_cancel_when_queued(self):
        return self.data.get("cancel")

    @property
    def cancel_message(self):
        return f"Delete instance not supported for computer {self.data.get('title')}"

    def _data(self, child_computer):
        return {
            "id": child_computer and child_computer.id,
            "title": child_computer and child_computer.title,
            "cancel": not (child_computer and child_computer.delete_instance_supported),
        }


@requires_advisor
class SetDefaultChildComputerActivity(GRPCActivityMixin, AbstractActivityInfo):
    """
    Activity to put a message on the queue for a hostagent to set a child
    computer instance as the default.
    """

    @requires("ViewActivity")
    def get_summary(self):
        return f"Set instance {self.data.get('title')} as default"

    @property
    def _message(self):
        return json.dumps({"set_default": self.data.get("title")})

    @property
    def should_cancel_when_queued(self):
        return self.data.get("cancel")

    @property
    def cancel_message(self):
        return (
            f"Set default instance not supported for computer {self.data.get('title')}"
        )

    def _data(self, child_computer):
        return {
            "id": child_computer and child_computer.id,
            "title": child_computer and child_computer.title,
            "cancel": not (
                child_computer and child_computer.set_default_instance_supported
            ),
        }


@requires_advisor
class ShutdownHostComputerActivity(GRPCActivityMixin, AbstractActivityInfo):
    """
    Activity to put a message on the queue for a hostagent to shutdown a
    host computer.
    """

    def __init__(self, activity, computer):
        self.activity = activity
        self.data = self._data(computer)
        self._set_activity_entity(computer.get_account_twin())
        self.activity_id = activity.id
        self.account_id = activity.account_id
        self.queue = self.computer.secure_id.decode()
        self.message = self._message

    @requires("ViewActivity")
    def get_summary(self):
        return f"Shutdown host {self.data.get('title')}"

    @property
    def _message(self):
        return json.dumps({"shutdown_host": self.data.get("id")})

    @property
    def should_cancel_when_queued(self):
        return self.data.get("cancel")

    @property
    def cancel_message(self):
        return f"Shutdown host not supported for computer {self.data.get('title')}"

    def _data(self, computer):
        return {
            "id": computer.id,
            "title": computer.title,
            "cancel": not computer.shutdown_host_supported,
        }


@requires_advisor
class StartChildComputerActivity(GRPCActivityMixin, AbstractActivityInfo):
    """
    Activity to put a message on the queue for a hostagent to start a child
    computer instance.
    """

    @requires("ViewActivity")
    def get_summary(self):
        return f"Start instance {self.data.get('title')}"

    @property
    def _message(self):
        return json.dumps({"start": self.data.get("id")})

    @property
    def should_cancel_when_queued(self):
        return self.data.get("cancel")

    @property
    def cancel_message(self):
        return f"Start instance not supported for computer {self.data.get('title')}"

    def _data(self, child_computer):
        return {
            "id": child_computer and child_computer.title,
            "title": child_computer and child_computer.title,
            "cancel": not (child_computer and child_computer.start_instance_supported),
        }


@requires_advisor
class StopChildComputerActivity(GRPCActivityMixin, AbstractActivityInfo):
    """
    Activity to put a message on the queue for a hostagent to stop a child
    computer instance.
    """

    def __init__(self, activity, computer, child_computer=None):
        self.activity = activity
        self.data = self._data(computer, child_computer)
        if computer:
            self._set_activity_entity(computer.get_account_twin())
        else:
            self._set_activity_entity(child_computer.get_account_twin())
        self.activity_id = activity.id
        self.account_id = activity.account_id

    @property
    def queue(self):
        return self.data.get("queue")

    @requires("ViewActivity")
    def get_summary(self):
        return f"Stop instance {self.data.get('title')}"

    @property
    def message(self):
        return self._message

    @property
    def _message(self):
        return json.dumps({"stop": self.data.get("id")})

    @property
    def should_cancel_when_queued(self):
        return self.data.get("cancel")

    @property
    def cancel_message(self):
        return f"Stop instance not supported for computer {self.data.get('title')}"

    def _data(self, computer, child_computer):
        return {
            "id": child_computer and child_computer.title,
            "title": child_computer and child_computer.title,
            "cancel": not (child_computer and child_computer.stop_instance_supported),
            "queue": computer.secure_id.decode() if computer else "",
        }


class WrongActivityChildError(Exception):
    """The activity child is the wrong one for this info class"""


class ChildComputerKeepAliveMixin:
    """
    Mixin for a ComputerRequest that needs keep-alive messaging for
    child computer instances.

    For legacy activity classes, this mixin should be inherited before
    the ComputerRequest class:

    class SomeLegacyActivity(ChildComputerKeepAliveMixin, ComputerRequest):
        ...

    For modern activity classes, this mixin should be inherited before both
    ComputerRequestMixin and AbstractActivityInfo

    class SomeModernActivity(
        ChildComputerKeepAliveMixin, ComputerRequestMixin, AbstractActivityInfo
    ):
        ...

    The following will automatically happen for any child activity with this mixin
    when it is added to a parent activity:
     - add a StartChildComputerActivity to the parent activity for any computer
       for which both must_keep_instance_alive and start_instance_supported
       are True
     - add a StopChildComputerActivity to the parent activity for any computer
       for which both must_keep_instance_alive and start_instance_supported
       are True
     - set the StopChildComputerActivity to be after the original child activity
    """

    @property
    @requires("ViewComputerPackages")
    def must_keep_instance_alive(self):
        """
        Whether or not the associated computer is a child instance that
        must be kept running until the activity reaches a status in
        FINAL_STATUSES.
        """
        return self.computer.is_wsl_instance

    def add_child_computer_activities(self, activity_parent, activity_child):
        if activity_child.info is not self:
            raise WrongActivityChildError(
                "The child activity provided is not the right one "
                "for this activity info class."
            )
        if self.must_keep_instance_alive:
            self.add_start_child_computer_activity(activity_parent, activity_child)
            self.add_stop_child_computer_activity(activity_parent, activity_child)

    def add_start_child_computer_activity(self, activity_parent, activity_child):
        if activity_child.info is not self:
            raise WrongActivityChildError(
                "The child activity provided is not the right one "
                "for this activity info class."
            )
        computer_child = activity_child.info.computer
        computer_parent = computer_child.get_parent_computer()
        if computer_child.start_instance_supported:
            start_child_computer_activity = activity_parent.add(
                StartChildComputerActivity,
                computer=computer_parent,
                child_computer=computer_child,
            )
            EventLog(activity_parent.account, activity_parent.creator).create_log_entry(
                computer_parent,
                "Activity created: " + start_child_computer_activity.info.get_summary(),
            )

    def add_stop_child_computer_activity(self, activity_parent, activity_child):
        if activity_child.info is not self:
            raise WrongActivityChildError(
                "The child activity provided is not the right one "
                "for this activity info class."
            )
        computer_child = activity_child.info.computer
        computer_parent = computer_child.get_parent_computer()
        if computer_child.start_instance_supported:
            stop_child_computer_activity = activity_parent.add(
                StopChildComputerActivity,
                computer=computer_parent,
                child_computer=computer_child,
            )
            stop_child_computer_activity.after(
                activity_child, minimum_progress=100, requires_success=False
            )
            EventLog(activity_parent.account, activity_parent.creator).create_log_entry(
                computer_parent,
                "Activity created: " + stop_child_computer_activity.info.get_summary(),
            )
