# Copyright 2005-2023 Canonical Limited. All rights reserved.

import re

from collections import defaultdict
from typing import Optional


HARDWARE_DEINDEXER_REGEX = re.compile("(:\d+)(?=\.)")
NOT_AVAILABLE = "Not available"


def deindex_hardware(items):
    """Remove any device indices from a single record of hardware info."""
    key = items[0]
    rest = items[1:]
    deindexed_key = HARDWARE_DEINDEXER_REGEX.sub("", key)
    return (deindexed_key,) + rest


def get_devices_info(hardware_info: list, device_key: str) -> list:
    """Return info for devices with the specified key."""
    usb_info = []

    for device in group_by_resource_key(hardware_info, device_key):
        vendor = get_value(device, "vendor")
        model = get_value(device, "product")
        description = get_value(device, "description")

        if not (
            vendor == NOT_AVAILABLE
            and model == NOT_AVAILABLE
            and description == NOT_AVAILABLE
        ):
            data = {"description": description, "model": model, "vendor": vendor}
            usb_info.append(data)

    return usb_info


def get_value(
    hardware_info: list,
    key: str,
    default: str = NOT_AVAILABLE,
    start: bool = True,
    ignore: Optional[list] = None,
    int_value: bool = False,
) -> str:
    """
    Extracts hardware information using the provided key or return a
    default.

    @param hardware_info: A list containing hardware information.
    @param key: The key to be used to retrieve the information.
    @param default: A default value to be returned if no key is found.
    @param start: If the match must happen at the beginning of the string.
    @param ignore: The keys to ignore.
    @param int_value: Use the int_value or not.
    """
    key = group_by_key(hardware_info, key, start=start, ignore=ignore)

    if not key:
        return default

    # Multiple keys may match, use the first one
    key = key[0]

    return key[2] if int_value else key[1]


def group_by_key(
    hardware_info: list,
    key: str,
    start: bool = False,
    ignore: Optional[list] = None,
) -> list:
    """
    Loops over the hardware_info list and collect keys containing the passed
    in key.

    @param hardware_info: A list containing hardware information in tuples.
    @param key: The key to be used as a filter.
    @param start: Perform or not an exact match at the beginning of the key.
    @param ignore: A list containing keys to be excluded from the final
        result.
    """
    if ignore is not None and not isinstance(ignore, list):
        ignore_keys = [ignore]
    else:
        ignore_keys = ignore

    if start:
        # Checks for matches at the beginning of string and then remove the
        # ones that should be ignored.
        keys = [info for info in hardware_info if info[0].startswith(key)]
    else:
        # Checks for matches in the key string and remove the ones to be
        # ignored.
        keys = [info for info in hardware_info if key in info[0]]

    return remove_keys(keys, ignore_keys) if ignore_keys else keys


def group_by_resource_key(
    hardware_info: list,
    resource_key: str,
    ignore: Optional[list] = None,
    start: bool = False,
) -> list:
    """Group hardware info based on the provided resource key.

    Return a list of lists of tuples from C{hardware_info}, grouped by key.

    Each group contains hardware info whose keys are equal up to the
    provided C{resource_key}, taking into account also indices that follow
    it.

    For example, if resource_key="bar", info with keys "foo.bar:0.baz" and
    "foo.bar:0.bza" will go in the same, while "foo.bar:0.baz" and
    "foo.bar:1.bza" will go in different groups because of the index.

    Note that keys prefixes up to the C{resource_key} (and index if present)
    are stripped from the result, so "foo.bar:0.baz" -> "baz".

    @param ignore: If specified, a list of key tokens that will be ignored.
    @param start: If true, match the C{resource_key} at the beginning of the
        info keys.
    """
    # Escape dots since the key is used in a regexp
    resource_key = resource_key.replace(".", r"\.")
    prefix_exp = "" if start else r".*\b"
    regexp = re.compile(r"(%s%s(:[0-9]+)?\.)" % (prefix_exp, resource_key))

    if ignore is not None:
        ignore = set(ignore)

    resource_info = defaultdict(list)

    for info in hardware_info:
        key = info[0]

        if ignore is not None and ignore.intersection(split_key(key)):
            # The key contains at least one element to ignore, skip it
            continue

        match = regexp.match(key)

        if match:
            prefix = match.groups()[0]

            # Strip the common prefix
            stripped_info = (key[len(prefix) :], info[1], info[2])
            resource_info[prefix].append(stripped_info)

    # Sort items based on the original prefix, in case indices were present.
    return [info_list for _, info_list in sorted(resource_info.items())]


def remove_keys(keys: list, ignore: list) -> list:
    """
    It looks tricky but it's not.

    It takes the key, replaces occurrances of ":" by ".", breaks it by ".",
    creates a set and check the intersection with the ignore key list and
    returns it.
    """
    return [key for key in keys if not set(split_key(key[0])).intersection(ignore)]


def split_key(key: str) -> list:
    """Replaces occurrances of ":" by "." and breaks it by "."."""
    return key.replace(":", ".").split(".")
