# Copyright 2005-2022 Canonical Limited. All rights reserved.

import csv

from io import StringIO

from canonical.content.download import set_csv_download

from canonical.landscape.model.resource.hardware import HardwareInfoManagement
from canonical.landscape.model.resource.hardware_info_parser import (
    HardwareInfoParser,
    NOT_AVAILABLE,
)
from canonical.landscape.ui.toolbar.page import ComputerPaginatorPage


class ComputerHardwareOverviewPage(ComputerPaginatorPage):
    title = "Hardware"

    def __init__(self, context, request):
        super().__init__(context, request)

        self.management = HardwareInfoManagement(self.account)
        hardware_info = self.management.get_hardware_for_computer_ids(
            [self.computer.id]
        )[self.computer.id]
        self.hardware = HardwareInfoParser(self.computer.id, hardware_info).hardware

    def _flatten_info(self, info, prefix=""):
        """
        Generator yielding (key, value) pairs with flattened hardware info.
        """
        if isinstance(info, dict):
            for key, value in info.items():
                _prefix = "%s.%s" % (prefix, key) if prefix else key
                for result in self._flatten_info(value, prefix=_prefix):
                    yield result
        elif isinstance(info, list):
            for index, value in enumerate(info):
                _prefix = "%s:%s" % (prefix, index) if prefix else index
                for result in self._flatten_info(value, prefix=_prefix):
                    yield result
        else:
            yield (prefix, info)

    def download_csv(self):
        """
        Return hardware information in CSV format for the computer selection.
        """
        set_csv_download(self.request.response, "hardware")
        computers_hardware = self.management.get_hardware_for_computer_ids(
            self.context.selection.get_selected_computer_ids()
        )

        column_headers = ("Computer Id", "Hardware Key", "Value")
        csv_data = [column_headers]

        for computer_id, hardware_info in computers_hardware.items():
            hardware = HardwareInfoParser(computer_id, hardware_info).hardware
            csv_data.extend(
                (computer_id, key, value)
                for key, value in self._flatten_info(hardware)
                if value != NOT_AVAILABLE
            )

        output = StringIO()
        writer = csv.writer(output)
        writer.writerows(csv_data)

        return output.getvalue()
