# Copyright 2005-2024 Canonical Limited. All rights reserved.

import os

from dataclasses import dataclass

from zope.cachedescriptors.property import Lazy
from zope.interface import implementer

from canonical.routes import Redirect
from canonical.security.protect import requires, requires_advisor, PUBLIC, PRIVATE

from canonical.landscape.api.error import (
    SecretNotFoundError,
    SecretsAuthenticationError,
)
from canonical.landscape.api.secrets_client import HTTPVaultClient
from canonical.landscape.model.configuration import is_standalone
from canonical.landscape.ui.account.page import AccountPage
from canonical.landscape.ui.secrets.interfaces import ISecrets


@requires_advisor
@implementer(ISecrets)
class Secrets:
    requires(PUBLIC, PRIVATE, ["account"])

    def __init__(self, account):
        self.account = account


@requires_advisor
class SecretsPage(AccountPage):
    """Base for the other secrets management pages."""

    requires(PUBLIC, PRIVATE, ["browserDefault"])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.account_url = self.context.account.url
        self.error_redirect_url = self.account_url

        if not is_standalone():
            raise Redirect(self.account_url)

    @Lazy
    def client(self):
        client = None

        try:
            client = HTTPVaultClient(self.context.account.name)
        except SecretsAuthenticationError as e:
            self.notes.error("Unable to reach secrets client")
            self.request.response.redirect(self.error_redirect_url)

        return client


class CreateSecretPage(SecretsPage):
    title = "Create Secret"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.error_redirect_url = f"{self.account_url}/secrets/create"

    def apply_changes(self):
        if self.request.method != "POST":
            return self()

        if not self.client:
            return

        secrets_name = self.request.form.get("secretsname")

        if not secrets_name:
            self.notes.error("Secrets name is a required field")
            self.request.response.redirect(self.error_redirect_url)

            return

        # Error if this secret already exists
        if self.client.list_secrets(secrets_name):
            self.notes.error(f"Secrets name {secrets_name} already exists")
            self.request.response.redirect(self.error_redirect_url)

            return

        for key, value in self.request.form.items():
            fields = key.split("_")

            if len(fields) < 2 or not value:
                continue

            field, n = fields[:2]
            secret_n = self.request.form.get(f"secret_{n}")

            if field != "dataid" or not secret_n:
                continue

            path = os.path.join(secrets_name, value)
            secret = {value: secret_n}

            try:
                self.client.create_or_update_secret(path, secret)
            except Exception:
                self.notes.error(f"Secret with data id {value} was not created")

        self.notes.info("Secret created")
        self.request.response.redirect(f"{self.account_url}/secrets")


class ListSecretsPage(SecretsPage):
    title = "Secrets"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.has_secrets = self.client is not None

    def apply_changes(self):
        if self.request.method != "POST":
            return self()

        if not self.client:
            return

        secrets_names = self.request.form.get("secretsname")

        if not secrets_names:
            self.notes.error("At least one secret must be selected")
            self.request.response.redirect(self.error_redirect_url)

            return

        if isinstance(secrets_names, str):
            secrets_names = [secrets_names]

        for secrets_name in secrets_names:
            for data_id in self.client.list_secrets(secrets_name):
                path = os.path.join(secrets_name, data_id)
                self.client.delete_secret(path)

        self.notes.info("Secret(s) removed")
        self.request.response.redirect(f"{self.account_url}/secrets")

    @Lazy
    def secrets(self):
        try:
            secrets = self.client.list_secrets()
        except Exception:
            secrets = []

        return secrets


@dataclass
class SecretKV:
    key: str
    value: str


class ViewSecretPage(SecretsPage):
    """
    XXX: view secret is not used at the moment, but may be added in the future.
    """

    title = "View Secret"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.data = {}
        self.secret_exists = False

        if not self.client:
            return

        secrets_name = self.request.form.get("secretsname")

        try:
            data = self.client.list_secrets(secrets_name)

            for datum in data:
                datum_path = os.path.join(secrets_name, datum)
                secret = self.client.read_secret(datum_path, datum)
                self.data[datum] = secret["data"][datum]

            self.secret_exists = True
        except SecretNotFoundError as e:
            self.notes.error(f"Unable to retrieve secret '{secrets_name}'")

    @Lazy
    def secret(self):
        kv_pairs = [SecretKV(key=key, value=value) for key, value in self.data.items()]

        return kv_pairs
