# Copyright 2005-2023 Canonical Limited. All rights reserved.
"""An internal representation of the language of reprepro listfilter conditions."""

from abc import ABC, abstractmethod


class RepreproFilter(ABC):
    """Base class for conditions suitable for use with reprepro commands that support filtering.

    See the reprepro manual section for 'listfilter' for more information. This base class acts as a
    dummy and a default that applies no filtering.
    """

    def __and__(self, other: "RepreproFilter") -> "RepreproFilter":
        return AndFilter(self, other)

    def __or__(self, other: "RepreproFilter") -> "RepreproFilter":
        return OrFilter(self, other)

    def __invert__(self) -> "RepreproFilter":
        return NotFilter(self)

    @abstractmethod
    def __str__(self) -> str:
        """Serializes this filter as a string that can be passed as the 'condition' to
        'reprepro listfilter'.
        """
        ...


class NoFilter(RepreproFilter):
    """A filter that imposes no filtering. Generally useful as a default."""

    def __str__(self):
        return ""


class CompoundFilter(RepreproFilter):
    """A filter that combines two or more reprepro filters.

    :param filters: The filters to be combined.
    """

    def __init__(self, *filters: RepreproFilter):
        assert (
            len(filters) >= 2
        ), "CompoundFilter subclasses can only be instantiated with at least two filters to combine"

        self.filters = filters

    def __str__(self) -> str:
        str_filters = self.operator.join(str(f) for f in self.filters)

        return f"({str_filters})"

    @property
    @abstractmethod
    def operator(self) -> str:
        ...


class AndFilter(CompoundFilter):
    """A conjuntion of two reprepro filters.

    :param filters: The filters to combine.
    """

    operator = ", "


class OrFilter(CompoundFilter):
    """A disjunction of two reprepro filters.

    :param filters: The filters to combine.
    """

    operator = " | "


class NotFilter(RepreproFilter):
    """An inverted reprepro filter.

    :param filter: The filter to invert.
    """

    def __init__(self, filter: RepreproFilter):
        self.filter = filter

    def __str__(self) -> str:
        return f"!({str(self.filter)})"


class EqualFilter(RepreproFilter):
    """A filter that asserts that a field equals an argument exactly.

    :param value: The value the field must equal.
    """

    def __init__(self, value: str):
        self.value = value

    def __str__(self) -> str:
        return f"{self.field} (== {self.value})"

    @property
    @abstractmethod
    def field(self) -> str:
        ...


class ArchitectureFilter(EqualFilter):
    """A filter on the CPU architecture of the packages."""

    field = "$Architecture"


class ComponentFilter(EqualFilter):
    """A filter on the CPU architecture of the packages."""

    field = "$Component"


class PackageTypeFilter(EqualFilter):
    """A filter on the CPU architecture of the packages."""

    field = "$PackageType"


class SourceFilter(EqualFilter):
    """A filter on the source package name."""

    field = "$Source"


class SourceVersionFilter(EqualFilter):
    """A filter on the source package name."""

    field = "$SourceVersion"


class VersionFilter(EqualFilter):
    """A filter on the Debian version number of the packages."""

    field = "$Version"


class PackageFilter(RepreproFilter):
    """A filter on the names of the packages.

    :param name: Filter only for packages containing this string.
    """

    def __init__(self, name: str):
        self.name = name

    def __str__(self) -> str:
        return f"Package (% *{self.name}*)"
