import os
from datetime import datetime, timedelta
from typing import Union
from uuid import uuid4

from dateutil.tz import tzutc

from twisted.internet.defer import Deferred, maybeDeferred
from twisted.logger import Logger
from twisted.python.reflect import safe_str
from twisted.web.resource import Resource
from twisted.web.server import Request, NOT_DONE_YET

from ..auth import signing_text
from ..credentials import Credentials
from ..service import ServiceEndpoint

from .call import Call
from .exception import APIError
from .schema import Date, Integer, RawStr, Schema, Unicode


log = Logger()


class QueryAPI(Resource):
    """Base class for query APIs.

    :param Registry registry: The registry used to look up `Method`s for handling API requests.
    :param str path: Optionally, the actual resource path the clients use when sending requests to
        this API. This is taken into account during signature validation. It can differ from the one
        in the request if the service is behind a reverse proxy. For this to work, you must ensure
        that 'path + path_of_the_rewritten_request' equals the resource path the clients are
        requesting.

    The following class attributes must be defined by sub-classes:

    :attr List[str] signature_versions: Allowed values for 'SignatureVersion'.
    :attr str content_type: The value for the 'Content-Type' HTTP header.
    """

    isLeaf = True
    time_format = "%Y-%m-%dT%H:%M:%SZ"

    schema = Schema(
        Unicode("Action"),
        RawStr("AccessKeyId"),
        Date("Timestamp", optional=True),
        Date("Expires", optional=True),
        RawStr("Version", optional=True),
        Unicode("SignatureMethod", optional=True, default="HmacSHA256"),
        Unicode("Signature"),
        Integer("SignatureVersion", optional=True, default=2),
    )

    def __init__(self, registry=None, path: Union[str, bytes] = ""):
        super(QueryAPI, self).__init__()

        self.registry = registry
        self.path: str

        if isinstance(path, bytes):
            self.path = path.decode()
        else:
            self.path = path

        #: The default validators used on each request.
        self._extra_validators = [self._validate_signature]

    @staticmethod
    def get_utc_time():
        """Returns the current C{datetime} in UTC."""
        return datetime.now(tzutc())

    def authorize(self, method, call):
        """Authorizes to invoke C{method} with C{call}."""

    def dump_error(self, error, request):
        """Serializes an error generating the response to send to the client.

        @param error: The L{APIError} to format.
        @param request: The request that generated the error.
        """
        raise NotImplementedError("Must be implemented by subclass")

    def dump_result(self, result):
        """Serializes the result of the method invocation.

        @param result: The L{Method} result to serialize.
        """
        return result

    def execute(self, call):
        """Execute an API L{Call}.

        Precondition: The request is authenticated; C{call.principal} is set to
        the L{Principal} for the L{User} making the request.

        @return: The response to write into the request for C{call}.
        @raises: L{APIError} if execution fails, with appropriate HTTP status
            code.
        """
        method = self.get_method(call)
        deferred = maybeDeferred(self.authorize, method, call)
        deferred.addCallback(lambda _: method.invoke(call))

        return deferred.addCallback(self.dump_result)

    def get_call_arguments(self, request):
        """
        Gets call arguments from a request. Override this is you want to use a
        different wire format.

        The return value is a dict with keys 'transport_args', 'handler_args',
        and 'raw_args'.

        The value of 'transport_args' must be a dict with these keys:
          - action
          - access_key_id
          - timestamp
          - expires
          - version
          - signature_method
          - signature
          - signature_version

        The value of 'handler_args' should be the application args that are
        passed to the action handler.

        The value of 'raw_args', the unprocessed args, are used for sig
        verification. This should be the same dict of data that the client used
        to sign the request. N.b., this data must not contain the sig itself.
        """
        params = {k: v[-1] for k, v in request.args.items()}
        args, rest = self.schema.extract(params)
        # Get rid of Signature so it doesn't mess with sig verification.
        params.pop("Signature", "")
        return {
            "transport_args": {
                "action": args.Action,
                "access_key_id": args.AccessKeyId.decode(),
                "timestamp": args.Timestamp,
                "expires": args.Expires,
                "version": args.Version.decode(),
                "signature_method": args.SignatureMethod,
                "signature": args.Signature,
                "signature_version": args.SignatureVersion,
            },
            "handler_args": rest,
            "raw_args": params,
        }

    def get_method(self, call, *args, **kwargs):
        """Return the L{Method} instance to invoke the given L{Call}.

        @param args: Positional arguments to pass to the method constructor.
        @param kwargs: Keyword arguments to pass to the method constructor.
        """
        method_class = self.registry.get(call.action, call.version)
        method = method_class(*args, **kwargs)

        if not method.is_available():
            raise APIError(
                400,
                "InvalidAction",
                "The action {} is not valid "
                "for this web service.".format(call.action),
            )

        return method

    def get_principal(self, access_key: bytes, request: Request) -> any:
        """Returns the principal object by access key.

        The returned object must have `access_key` and `secret_key`
        attributes and if the authentication succeeds, it will be passed to the
        created `Call`.
        """
        raise NotImplementedError("Must be implement by subclasses")

    def get_status_text(self):
        """Get the bytes to return when a status check is made."""
        return b"Query API Service"

    def handle(self, request):
        """Handles an HTTP request for executing an API call.

        Authenticates the request, checking its signature, then calls the
        C{execute} method, passing it a L{Call} set with the principal for the
        authenticated user and the generic parameters extracted from the
        request.

        @param request: The L{HTTPRequest} to handle.
        """
        request.id = str(uuid4())
        deferred = maybeDeferred(self._validate, request)
        deferred.addCallback(self.execute)

        def write_response(response):
            request.setHeader("Content-Length", str(len(response)))
            request.setHeader("Content-Type", self.content_type)
            # Prevent browsers from trying to guess a different content type.
            request.setHeader("X-Content-Type-Options", "nosniff")
            request.write(response.encode())
            request.finish()

            return response

        def write_error(failure):
            if failure.check(APIError):
                status = failure.value.status

                # Don't log tracebacks for 4xx responses.
                if status < 400 or status >= 500:
                    log.failure("", failure)
                else:
                    log.info(
                        "status: {status} message: {message}",
                        status=status,
                        message=safe_str(failure.value),
                    )

                body = failure.value.response
                if body is None:
                    body = self.dump_error(failure.value, request)
            else:
                # If the error is generic (not an APIError), log the message,
                # but don't send it back to the client, as it could contain
                # sensitive information. Send a generic server error message
                # instead.
                log.failure("", failure)
                body = "Server error"
                status = 500

            request.setResponseCode(status)

            write_response(body)

        deferred.addCallback(write_response)
        deferred.addErrback(write_error)

        return deferred

    def render_GET(self, request):
        if not request.args:
            request.setHeader("Content-Type", "text/plain")
            return self.get_status_text()

        self.handle(request)

        return NOT_DONE_YET

    render_POST = render_GET

    def _validate(self, request: Request) -> Deferred:
        """Validates `Request` before executing it.

        The following conditions are checked:
          - The request contains all the generic parameter
          - The action specified in the request is a supported one
          - Any extra validator callables contained in `_extra_validators`.

        :returns Deferred: The validated `Call`, with its default arguments set and the principal
            of the accessing user.
        """
        call_arguments = self.get_call_arguments(request)
        args = call_arguments["transport_args"]
        rest = call_arguments["handler_args"]
        params = call_arguments["raw_args"]

        self._validate_generic_parameters(args)

        def create_call(principal):
            self._validate_principal(principal, args)

            for validator in self._extra_validators:
                validator(request, principal, args, params)

            call = Call(
                raw_params=rest,
                principal=principal,
                action=args["action"],
                version=args["version"],
                id=request.id,
                request=request,
            )

            return call

        deferred = maybeDeferred(self.get_principal, args["access_key_id"], request)
        deferred.addCallback(create_call)

        return deferred

    def _validate_generic_parameters(self, args):
        """Validates the generic request parameters.

        @param args: Parsed schema arguments.
        @raises APIError: In the following cases:
            - Action is not included in C{self.actions}
            - SignatureVersion is not included in C{self.signature_versions}
            - Expires and Timestamp are both present
            - Expires is before current time
            - Timestamp is older than 15 minutes
        """
        utc_now = self.get_utc_time()

        self.registry.check(args["action"], args["version"])

        if not args["signature_version"] in self.signature_versions:
            raise APIError(
                403,
                "InvalidSignature",
                "SignatureVersion '{}' not supported".format(args["signature_version"]),
            )

        if args["expires"] and args["timestamp"]:
            raise APIError(
                400,
                "InvalidParameterCombination",
                "The parameter Timestamp cannot be used with the parameter Expires",
            )

        if args["expires"] and args["expires"] < utc_now:
            raise APIError(
                400,
                "RequestExpired",
                "Request has expired. Expired at {}".format(
                    args["expires"].strftime(self.time_format)
                ),
            )

        if args["timestamp"] and args["timestamp"] + timedelta(minutes=15) < utc_now:
            raise APIError(
                400,
                "RequestExpired",
                "Request has expired. Timestamp is {}".format(
                    args["timestamp"].strftime(self.time_format)
                ),
            )

    def _validate_principal(self, principal, args):
        if principal is None:
            raise APIError(
                401,
                "AuthFailure",
                "No user with access key '{}'".format(args["access_key_id"]),
            )

    def _validate_signature(self, request, principal, args, params):
        signature_error = APIError(
            403,
            "SignatureDoesNotMatch",
            "The request signature we calculated does not match the signature you provided. Check your key and signing method.",
        )

        if "signature" not in args:
            raise signature_error

        creds = Credentials(principal.access_key, principal.secret_key)
        endpoint = ServiceEndpoint()
        endpoint.method = request.method.decode()
        endpoint.set_canonical_host(request.getHeader("Host"))

        path = request.path
        try:  # This is str in tests but bytes in general
            path = path.decode()
        except AttributeError:
            pass

        if not path.startswith(self.path):
            path = "{}/{}".format(self.path.rstrip("/"), path.lstrip("/"))

        endpoint.path = path
        signature = creds.sign(
            signing_text(endpoint, params), hash_type=args["signature_method"]
        )

        if signature != args["signature"]:
            raise signature_error
