# Copyright 2005-2024 Canonical Limited. All rights reserved.
import json
import io
from urllib.parse import urlencode
from unittest import mock

import transaction
from storm.twisted.testing import FakeTransactor
from zope.security.management import getSecurityPolicy, setSecurityPolicy
from zope.security.simplepolicies import PermissiveSecurityPolicy
from twisted.web.http import Request, parse_qs

from canonical.landscape.testing.model import TwistedModelTest
from canonical.landscape.api.helpers import ApiRequestError
from canonical.landscape.api.principal import Principal
from canonical.landscape.api.resource import ServiceAPI
from canonical.landscape.model.gpg.management import GPGKeyManagement
from canonical.landscape.model.repository.management import RepositoryManagement
from canonical.landscape.model.resource.hardware import HardwareInfoManagement
from canonical.landscape.model.account.savedsearch import SavedSearchManagement
from canonical.landscape.model.package.profile import PackageProfilesManagement
from canonical.txapi.server.exception import APIError
from canonical.txapi.server.registry import Registry


class APITestMixin:
    """
    Helper class with common tests for API methods classes.

    Tests in this class will be executed for subclasses of L{APITest}
    with the proper API method.
    """

    def test_registration(self):
        """
        The method is registered in the global API methods.
        """
        self.assertTrue(self.method.__venusian_callbacks__)

    def test_permission(self):
        """
        The method may only be invoked by users with the C{api_method_requires}
        permission.
        """
        if getattr(self.method, "requires", None) is None:
            # XXX new-style methods don't have the requires attribute
            return
        self.assertEqual(self.api_method_requires, self.method.requires)

    def test_authorize_method(self):
        """
        The L{Method} provides an C{authorize} method to check authorization.
        """
        self.assertTrue(callable(self.method.authorize))


class APITestMeta(type):
    """
    Metaclass to setup L{APITest} class.
    """

    def __new__(cls, name, bases, clsdict):
        if clsdict.get("api_method_class") is not None:
            bases += (APITestMixin,)

        return type.__new__(cls, name, bases, clsdict)


class APITest(TwistedModelTest, metaclass=APITestMeta):
    """
    A test class for an API L{Method}.

    @cvar api_method_class: the L{Method} class under test.
    """

    api_method_class = None
    api_method_requires = None

    with_person = True
    with_account = True

    with_zcml = True

    with_gpg_key_management = False
    with_repository_management = False
    with_hardware_management = False
    with_saved_search_management = False
    with_maas_management = False
    with_package_profile_management = False

    def setUp(self):
        super(APITest, self).setUp()

        if self.api_method_class is not None:
            self.credentials = self.person.create_api_credentials(self.account)
            self.principal = Principal(self.credentials)
            self.transactor = FakeTransactor(transaction)
            self.method = self.api_method_class(self.transactor)

        if self.with_gpg_key_management:
            self.gpg_key_management = GPGKeyManagement(self.account, self.person)
        if self.with_repository_management:
            self.repository_management = RepositoryManagement(self.account, self.person)
        if self.with_hardware_management:
            self.hardware_management = HardwareInfoManagement(self.account)
        if self.with_saved_search_management:
            self.saved_search_management = SavedSearchManagement(
                self.account, self.person
            )
        if self.with_package_profile_management:
            self.package_profile_management = PackageProfilesManagement(
                self.account, self.person
            )

    def assertAPIError(self, deferred, code, message, status=400):
        """
        Check that the call raises and L{APIError} with the specified status,
        code and message.

        @param deferred: a deferred returning the result
        @param code: the expected error code name.
        @param message: the expected error message.
        @param status: the expected response status.
        """

        def check(error):
            self.assertEqual(status, error.status)
            self.assertEqual(code, error.code)
            self.assertEqual(message, error.message)
            return error

        return self.assertFailure(deferred, APIError).addCallback(check)


class APIHandlerTest(TwistedModelTest):
    with_account = True
    with_person = True
    with_hardware_management = False

    def setUp(self):
        super().setUp()
        if self.with_hardware_management:
            self.hardware_management = HardwareInfoManagement(self.account)

        jwt_auth_mock = mock.patch(
            "canonical.landscape.api.jwt_auth.jwt_authenticate_request"
        ).start()
        jwt_auth_mock.return_value = mock.Mock(
            spec=Principal,
            account_id=self.account.id,
            id=self.person.id,
            account=self.account,
            person=self.person,
        )

        self.addCleanup(mock.patch.stopall)
        self.addCleanup(setSecurityPolicy, getSecurityPolicy())

        setSecurityPolicy(PermissiveSecurityPolicy)


class APIFromUrlTest(APIHandlerTest):
    def setUp(self):
        super().setUp()
        self.transactor = FakeTransactor()
        self.registry = Registry()
        self.api = ServiceAPI(self.transactor, registry=self.registry)

    async def send_request(self, path=None, method=None, body_args={}, query_args={}):
        if "?" in path:
            path, query = path.split("?")
        else:
            query = urlencode(query_args)
        query_args = parse_qs(query.encode())  # Parse using twisted which gives bytes

        request = mock.Mock(
            spec_set=Request,
            content=io.BytesIO(json.dumps(body_args).encode()),
            path=path.encode(),
            method=method.encode(),
            args=query_args,
        )

        try:
            await self.api.render_method_helper(request)
        except ApiRequestError:
            pass

        code_mock = request.setResponseCode
        code_mock.assert_called_once()
        write_mock = request.write
        write_mock.assert_called_once()

        response = json.loads(write_mock.call_args[0][0])
        return response, code_mock.call_args[0][0]
